import typing as t
from io import IOBase

from click import Abort
from tqdm import tqdm


if t.TYPE_CHECKING:
    import typing_extensions as te


T = t.TypeVar("T")


@t.overload
def invert(b: None) -> None:
    ...


@t.overload
def invert(b: "te.Literal[True]") -> "te.Literal[False]":
    ...


@t.overload
def invert(b: "te.Literal[False]") -> "te.Literal[True]":
    ...


@t.overload
def invert(b: t.Optional[bool]) -> t.Optional[bool]:
    ...


def invert(b: t.Optional[bool]):
    """Invert a `bool`, ignoring `None` values."""
    if b is None:
        return None
    return not b


def partition(predicate: t.Callable[[T], bool], iterable: t.Iterable[T]):
    """Partition a list based on the results of a :param:`predicate`."""
    trues: t.List[T] = []
    falses: t.List[T] = []
    for item in iterable:
        if predicate(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses


def multi_partition(*predicates: t.Callable[[T], bool], iterable: t.Iterable[T]):
    """Partition a list based on a series of :param:`predicates`.

    Predicates are checked in the order they are passed."""
    results: t.List[t.List[T]] = [[] for _ in predicates]
    results.append([])

    for item in iterable:
        i = 0
        matched = False
        for pred in predicates:
            if pred(item):
                results[i].append(item)
                matched = True
                break
            i += 1
        if not matched:
            results[-1].append(item)

    return tuple(results)


def tryExec(func: t.Callable[..., t.Any], *params: t.Any):
    try:
        func(*params)
    except:
        pass


def find(iter: t.Iterable[T], matches: t.Iterable[T]):
    return next((match for match in iter if match in matches), None)


_download_interrupt = False


def read_with_progress(
    input: IOBase,
    output: IOBase,
    size=0,
    blocksize=4096,
    label: t.Optional[str] = "",
    clear_progress=False,
):
    with tqdm(
        total=size,
        desc=label,
        leave=(not clear_progress),
        unit_scale=True,
        unit="b",
        delay=0.4,
        disable=False,
    ) as bar:
        while True:
            if _download_interrupt:
                raise Abort

            buf = input.read(blocksize)
            if not buf:
                break
            output.write(buf)
            bar.update(len(buf))


class GeneratorWithLen(t.Generic[T]):
    def __init__(self, gen: t.Iterator[T], length: int):
        self.gen = gen
        self.length = length

    def __iter__(self):
        return self.gen

    def __next__(self):
        return next(self.gen)

    def __len__(self):
        return self.length



from urllib.parse import urlparse

import click
import pytest

import mons.clickExt as clickExt


def color_cmd():
    @click.command()
    @clickExt.color_option()
    def cmd(color):
        return color

    return cmd


@pytest.mark.parametrize(
    ("color_arg", "output"),
    [
        ("auto", None),
        ("always", True),
        ("never", False),
    ],
)
def test_color_option(runner, color_arg, output):
    result = runner.invoke(color_cmd(), ["--color", color_arg], standalone_mode=False)
    assert result.exit_code == 0, result.output
    assert output == result.return_value


@pytest.mark.parametrize(("env", "output"), [({}, None), ({"NO_COLOR": "1"}, False)])
def test_color_option_auto(runner, env, output):
    result = runner.invoke(color_cmd(), env=env, standalone_mode=False)
    assert result.exit_code == 0, result.output
    assert output == result.return_value


def test_color_option_bad(runner):
    result = runner.invoke(color_cmd(), ["--color", "invalid"], standalone_mode=False)
    assert isinstance(result.exception, click.BadParameter)
    assert "Possible values: auto, never, always" in result.exception.message


@pytest.mark.parametrize(
    ("url_arg", "expect"),
    [
        ("https://mons.coloursofnoise.ca/file", "{url_arg}"),
        ("mons.coloursofnoise.ca/file", "file://{url_arg}"),
    ],
)
def test_url_type(runner, url_arg, expect):
    cmd = click.Command(
        "cmd",
        params=[click.Argument(["url"], type=clickExt.URL(default_scheme="file"))],
        callback=lambda url: url,
    )
    result = runner.invoke(cmd, [url_arg], standalone_mode=False)
    assert not result.exception, result.output
    assert result.return_value == urlparse(expect.format(url_arg=url_arg))


@pytest.mark.parametrize(
    ("param_args", "err"),
    [
        ({"valid_schemes": ["file"]}, "URI scheme 'https' not allowed"),
        ({"require_path": True}, "Path component required"),
    ],
)
def test_url_type_bad(runner, param_args, err):
    cmd = click.Command(
        "cmd", params=[click.Argument(["url"], type=clickExt.URL(**param_args))]
    )
    result = runner.invoke(cmd, ["https://mons.coloursofnoise.ca"])
    assert result.exception
    assert err in result.output


class TestCommandExt:
    @pytest.mark.parametrize(
        ("args", "expect"), [([], None), (["--opt"], True), (["--opt=value"], "value")]
    )
    def test_default_option(self, runner, args, expect):
        cmd = clickExt.CommandExt(
            "cmd",
            params=[
                clickExt.ExplicitOption(["--opt"]),
                clickExt.DefaultOption(["--opt"], is_flag=True),
            ],
            callback=lambda opt_default, opt: opt_default or opt,
        )

        result = runner.invoke(cmd, args, standalone_mode=False)
        assert result.return_value == expect

    def test_placeholder(self, runner):
        cmd = clickExt.CommandExt("cmd", params=[clickExt.PlaceHolder(["placeholder"])])

        result = runner.invoke(cmd, standalone_mode=False)
        assert not result.exception

    @pytest.mark.parametrize(
        ("default", "exception"),
        [
            (None, click.MissingParameter),
            (lambda: None, click.MissingParameter),
            ("A Value", None),
            (lambda: "A Value", None),
        ],
    )
    def test_optionalarg(self, runner, default, exception):
        cmd = clickExt.CommandExt(
            "cmd", params=[clickExt.OptionalArg(["arg"], default=default)]
        )

        result = runner.invoke(cmd, standalone_mode=False)
        assert exception is None or isinstance(result.exception, exception)



import typing as t

import pytest

from mons.mons import cli as mons_cli

if t.TYPE_CHECKING:
    from click.testing import CliRunner

TEST_INSTALL = "_mons_testing"


@pytest.mark.prioritize
@pytest.mark.must_pass
def test_add(runner: "CliRunner", test_install):
    result = runner.invoke(mons_cli, ["add", TEST_INSTALL, test_install])
    assert result.exit_code == 0, result.output
    assert "Found Celeste.exe" in result.output


def setup_install(runner: "CliRunner", test_install):
    result = runner.invoke(mons_cli, ["add", TEST_INSTALL, test_install])
    if result.exit_code != 0:
        pytest.skip(result.output)


def test_rename(runner, test_install):
    RENAMED_INSTALL = TEST_INSTALL + "_renamed"

    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["rename", TEST_INSTALL, RENAMED_INSTALL])
    assert result.exit_code == 0, result.output
    assert "Renamed install" in result.output


def test_set_path(runner, test_install):
    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["set-path", TEST_INSTALL, test_install])
    assert result.exit_code == 0, result.output
    assert "Found Celeste.exe" in result.output


def test_remove(runner: "CliRunner", test_install):
    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["remove", TEST_INSTALL, "--force"])
    assert result.exit_code == 0, result.output
    assert f"Removed install {TEST_INSTALL}" in result.output


def test_list(runner: "CliRunner", test_install):
    result = runner.invoke(mons_cli, ["list"])
    assert result.exit_code != 0
    assert "No installs found" in result.output

    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["list"])
    assert result.exit_code == 0, result.output
    assert TEST_INSTALL in result.output


def test_show(runner: "CliRunner", test_install):
    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["show", TEST_INSTALL])
    assert result.exit_code == 0, result.output
    assert TEST_INSTALL in result.output



import argparse
import os

import pytest

from mons.utils import find_celeste_file


def path_exists(path):
    if os.path.exists(path):
        try:
            return find_celeste_file(path, "Celeste.exe")
        except FileNotFoundError as e:
            raise argparse.ArgumentTypeError(e)
    raise argparse.ArgumentTypeError(f"Path {path} does not exist.")


def pytest_addoption(parser: pytest.Parser, pluginmanager):
    parser.addoption(
        "--mons-test-install", "--mons", type=path_exists, dest="mons_test_install"
    )


def pytest_configure(config: pytest.Config):
    config.addinivalue_line("markers", "prioritize: prioritize this test")
    config.addinivalue_line(
        "markers", "must_pass: this test must pass in order to continue"
    )


@pytest.hookimpl(tryfirst=True)
def pytest_collection_modifyitems(session, config, items):
    items.sort(key=lambda i: 0 if i.get_closest_marker("prioritize") else 1)


must_pass_failed = None


# https://stackoverflow.com/a/59392344
def pytest_runtest_makereport(item, call):
    global must_pass_failed
    if not must_pass_failed and item.iter_markers(name="must_pass"):
        if call.excinfo is not None:
            must_pass_failed = item


def pytest_runtest_setup(item):
    if must_pass_failed is not None:
        pytest.skip(f"must_pass test failed ({must_pass_failed.name})")



import typing as t

if t.TYPE_CHECKING:
    import click
import mons.clickExt as clickExt
from inspect import signature, unwrap


def unwrap_count(f):
    depth = 0

    def incr(_):
        nonlocal depth
        depth = depth + 1
        return False

    return unwrap(f, stop=incr), depth


def test_command_arguments(command: "click.Command"):
    assert command.callback

    callback, depth = unwrap_count(command.callback)

    callback_args = signature(callback).parameters.keys()
    callback_args_unique = set(callback_args)
    command_params = [
        param.name
        for param in command.params
        if param.expose_value and not isinstance(param, clickExt.PlaceHolder)
        if param.name
    ]

    for param in command_params:
        assert (
            param in callback_args
        ), f"Callback for '{command.name}' has missing argument: '{param}'"
        callback_args_unique.discard(param)

    assert depth == len(
        callback_args_unique
    ), f"Callback for '{command.name}' has too many positional arguments: '{callback_args_unique}"



import errno
import hashlib
import json
import os
import re
import time
import typing as t
import urllib.parse
import urllib.request
import zipfile
from contextlib import contextmanager
from urllib.error import HTTPError

import dnfile  # https://github.com/malwarefrank/dnfile
import urllib3
import yaml
from click import echo
from pefile import DIRECTORY_ENTRY  # https://github.com/erocarrera/pefile
from tqdm import tqdm

from mons.baseUtils import GeneratorWithLen
from mons.downloading import download_with_progress
from mons.modmeta import ModMeta
from mons.modmeta import read_mod_info


VANILLA_HASH: t.Dict[str, t.Tuple[str, str]] = {
    "f1c4967fa8f1f113858327590e274b69": ("1.4.0.0", "FNA"),
    "107cd146973f2c5ec9fb0b4f81c1588a": ("1.4.0.0", "XNA"),
}


@contextmanager
def timed_progress(msg: str):
    """Times execution of the current context, then prints :param:`msg` with :func:`tqdm.write`.

    :param msg: Message to be printed. Formatted with a `time` kwarg."""
    start = time.perf_counter()
    yield
    end = time.perf_counter()
    # Carriage return ensures msg is printed properly even after multiple progress bars
    tqdm.write("\r" + msg.format(time=end - start))


def find_celeste_file(path: str, file: str, force_name=True):
    if os.path.basename(path) == "Celeste.app":
        path = os.path.join(path, "Resources")

    ret = path
    if os.path.isdir(path):
        ret = os.path.join(path, file)
        if not os.path.exists(ret):
            raise FileNotFoundError(
                errno.ENOENT, f"File `{file}` could not be found in `{path}`", ret
            )
    elif force_name and not os.path.basename(path) == file:
        raise FileNotFoundError(errno.ENOENT, f"File `{file}` not found", path)
    return ret


def getMD5Hash(path: str):
    with open(path, "rb") as f:
        file_hash = hashlib.md5()
        chunk = f.read(8129)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(8129)
    return file_hash.hexdigest()


def unpack(zip: zipfile.ZipFile, root: str, prefix="", label="Extracting"):
    totalSize = 0
    for zipinfo in zip.infolist():
        if not prefix or zipinfo.filename.startswith(prefix):
            totalSize += zipinfo.file_size

    with tqdm(total=totalSize, desc=label, leave=False) as bar:
        for zipinfo in zip.infolist():
            if not zipinfo.filename or zipinfo.filename.endswith("/"):
                continue

            if prefix:
                if not zipinfo.filename.startswith(prefix):
                    continue
                zipinfo.filename = zipinfo.filename[len(prefix) :]

            zip.extract(zipinfo, root)
            bar.update(zipinfo.file_size)


class EverestHandler(urllib.request.BaseHandler):
    def everest_open(self, req: urllib.request.Request):
        parsed_url = urllib.parse.urlparse(req.full_url)
        gb_url = re.match("^(https://gamebanana.com/mmdl/.*),.*,.*$", parsed_url.path)
        download_url = gb_url[1] if gb_url else parsed_url.path
        req.full_url = download_url
        return self.parent.open(req)


opener = urllib.request.build_opener(EverestHandler)
opener.addheaders = [
    (
        "User-Agent",
        "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1941.0 Safari/537.36",
    )
]
urllib.request.install_opener(opener)


def parseExeInfo(path: str):
    echo("Reading exe...\r", nl=False)
    pe = dnfile.dnPE(path, fast_load=True)
    pe.parse_data_directories(
        directories=DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR"]
    )

    assert pe.net

    stringHeap: dnfile.stream.StringsHeap = pe.net.metadata.streams_list[1]  # type: ignore

    hasEverest = False
    everestBuild = None

    heapSize = stringHeap.sizeof()
    i = 0
    with tqdm(total=heapSize, desc="Scanning exe", leave=False) as bar:
        while i < len(stringHeap.__data__):
            string = stringHeap.get(i)
            if string is None:
                break
            string = str(string)
            if string == "EverestModule":
                hasEverest = True
            if string.startswith("EverestBuild"):
                everestBuild = string[len("EverestBuild") :]
                hasEverest = True
                break
            inc = max(len(string), 1)
            i += inc
            bar.update(inc)

    assemRef = pe.net.mdtables.AssemblyRef
    assert assemRef
    framework = "FNA" if any(row.Name == "FNA" for row in assemRef.rows) else "XNA"

    return hasEverest, everestBuild, framework


def parseVersionSpec(string: str):
    if string.startswith("1.") and string.endswith(".0"):
        string = string[2:-2]
    if string.isdigit():
        buildnumber = int(string)
    else:
        buildnumber = latest_build(string)

    return buildnumber


def latest_build(branch: str):
    if branch.startswith(("refs/heads/", "refs/pull/")):
        return latest_build_azure(branch)

    build_list = get_build_list()
    for build in build_list:
        if not branch or build["branch"] == branch:
            return int(build["version"])

    return None


def latest_build_azure(branch: str):
    response: urllib3.HTTPResponse = urllib3.PoolManager().request(
        "GET",
        "https://dev.azure.com/EverestAPI/Everest/_apis/build/builds",
        fields={
            "definitions": 3,
            "statusFilter": "completed",
            "resultFilter": "succeeded",
            "branchName": branch
            if branch == "" or branch.startswith(("refs/heads/", "refs/pull/"))
            else "refs/heads/" + branch,
            "api-version": 6.0,
            "$top": 1,
        },
    )

    data: t.Dict[str, t.Any] = json.loads(response.data.decode())
    if data["count"] < 1:
        return None
    elif data["count"] > 1:
        raise Exception("Unexpected number of builds: " + str(data["count"]))

    build = data["value"][0]
    id = build["id"]
    try:
        return int(id) + 700
    except:
        pass
    return None


def build_exists(build: int):
    build_list = get_build_list()
    if build in (int(b["version"]) for b in build_list):
        return True

    return build_exists_azure(build)


def build_exists_azure(build: int):
    try:
        urllib.request.urlopen(
            "https://dev.azure.com/EverestAPI/Everest/_apis/build/builds/"
            + str(build - 700)
        )
        return True
    except HTTPError as err:
        if err.code == 404:
            return False
        raise


updateURLLookup = {
    "main": "mainDownload",
    "olympus-meta": "olympusMetaDownload",
    "olympus-build": "olympusBuildDownload",
}


def fetch_build_artifact(build: int, artifactName: str) -> urllib3.HTTPResponse:
    build_list = get_build_list()
    for b in build_list:
        if build == int(b["version"]):
            return urllib3.PoolManager().request(
                "GET", b[updateURLLookup[artifactName]], preload_content=False
            )

    return fetch_build_artifact_azure(build, artifactName)


def fetch_build_artifact_azure(
    build: int, artifactName="olympus-build"
) -> urllib3.HTTPResponse:
    return urllib3.PoolManager().request(
        "GET",
        f"https://dev.azure.com/EverestAPI/Everest/_apis/build/builds/{build - 700}/artifacts",
        fields={
            "artifactName": artifactName,
            "api-version": 6.0,
            "$format": "zip",
        },
        preload_content=False,
    )


build_list = None


def get_build_list() -> t.List[t.Dict[str, t.Any]]:
    global build_list
    if build_list:
        return build_list

    update_url = (
        urllib.request.urlopen("https://everestapi.github.io/everestupdater.txt")
        .read()
        .decode()
        .strip()
    )

    build_list = yaml.safe_load(
        download_with_progress(update_url, None, "Downloading Build List", clear=True)
    )
    return build_list


mod_list = None


def get_mod_list() -> t.Dict[str, t.Any]:
    global mod_list
    if mod_list:
        return mod_list

    update_url = (
        urllib.request.urlopen("https://everestapi.github.io/modupdater.txt")
        .read()
        .decode()
        .strip()
    )
    mod_list = yaml.safe_load(
        download_with_progress(
            update_url,
            None,
            "Downloading Update List",
            clear=True,
        )
    )
    return mod_list


dependency_graph = None


def get_dependency_graph() -> t.Dict[str, t.Any]:
    global dependency_graph
    if dependency_graph:
        return dependency_graph

    dependency_graph = yaml.safe_load(
        download_with_progress(
            "https://max480-random-stuff.appspot.com/celeste/mod_dependency_graph.yaml?format=everestyaml",
            None,
            "Downloading Dependency Graph",
            clear=True,
        )
    )
    return dependency_graph


def search_mods(search: str):
    search = urllib.parse.quote_plus(search)
    url = (
        f"https://max480-random-stuff.appspot.com/celeste/gamebanana-search?q={search}"
    )
    response = urllib.request.urlopen(url)
    return yaml.safe_load(response.read())


def read_blacklist(path: str):
    with open(path) as file:
        return [m.strip() for m in file.readlines() if not m.startswith("#")]


def installed_mods(
    path: str,
    *,
    dirs: t.Optional[bool] = None,
    valid: t.Optional[bool] = None,
    blacklisted: t.Optional[bool] = None,
    with_size=False,
    with_hash=False,
) -> t.Iterator[ModMeta]:
    files = os.listdir(path)
    blacklist = None
    if os.path.isfile(os.path.join(path, "blacklist.txt")):
        blacklist = read_blacklist(os.path.join(path, "blacklist.txt"))
        if blacklisted is not None:
            files = list(filter(lambda m: blacklisted ^ (m in blacklist), files))
    elif blacklisted:
        files = []

    def _iter():
        for file in files:
            modpath = os.path.join(path, file)
            if dirs is not None:
                if dirs ^ bool(os.path.isdir(modpath)):
                    continue

            mod = read_mod_info(modpath, with_size=with_size, with_hash=with_hash)
            if valid is not None:
                if valid ^ bool(mod):
                    continue
            mod = mod or ModMeta.placeholder(modpath)
            if not mod:
                continue
            if blacklist and file in blacklist:
                mod.Blacklisted = True
            yield mod

    return GeneratorWithLen(_iter(), len(files))


def enable_mods(path: str, *mods: str):
    blacklist_path = os.path.join(path, "blacklist.txt")
    if os.path.isfile(blacklist_path):
        with open(blacklist_path) as file:
            blacklist = file.readlines()
        i = 0
        while i < len(blacklist):
            if blacklist[i].strip() in mods:
                blacklist[i] = "#" + blacklist[i]
            i += 1
        with open(blacklist_path, mode="w") as file:
            file.writelines(blacklist)



import typing as t

# https://stackoverflow.com/a/63839503
METRIC_LABELS: t.List[str] = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
BINARY_LABELS: t.List[str] = [
    "B",
    "KiB",
    "MiB",
    "GiB",
    "TiB",
    "PiB",
    "EiB",
    "ZiB",
    "YiB",
]
PRECISION_OFFSETS: t.List[float] = [0.5, 0.05, 0.005, 0.0005]  # PREDEFINED FOR SPEED.
PRECISION_FORMATS: t.List[str] = [
    "{}{:.0f} {}",
    "{}{:.1f} {}",
    "{}{:.2f} {}",
    "{}{:.3f} {}",
]  # PREDEFINED FOR SPEED.


def format_bytes(
    num: t.Union[int, float], metric: bool = False, precision: int = 1
) -> str:
    """
    Human-readable formatting of bytes, using binary (powers of 1024)
    or metric (powers of 1000) representation.
    """

    assert isinstance(num, (int, float)), "num must be an int or float"
    assert isinstance(metric, bool), "metric must be a bool"
    assert (
        isinstance(precision, int) and precision >= 0 and precision <= 3
    ), "precision must be an int (range 0-3)"

    unit_labels = METRIC_LABELS if metric else BINARY_LABELS
    last_label = unit_labels[-1]
    unit_step = 1000 if metric else 1024
    unit_step_thresh = unit_step - PRECISION_OFFSETS[precision]

    is_negative = num < 0
    if is_negative:  # Faster than ternary assignment or always running abs().
        num = abs(num)

    for unit in unit_labels:
        if num < unit_step_thresh:
            # VERY IMPORTANT:
            # Only accepts the CURRENT unit if we're BELOW the threshold where
            # float rounding behavior would place us into the NEXT unit: F.ex.
            # when rounding a float to 1 decimal, any number ">= 1023.95" will
            # be rounded to "1024.0". Obviously we don't want ugly output such
            # as "1024.0 KiB", since the proper term for that is "1.0 MiB".
            break
        if unit != last_label:
            # We only shrink the number if we HAVEN'T reached the last unit.
            # NOTE: These looped divisions accumulate floating point rounding
            # errors, but each new division pushes the rounding errors further
            # and further down in the decimals, so it doesn't matter at all.
            num /= unit_step

    return PRECISION_FORMATS[precision].format("-" if is_negative else "", num, unit)  # type: ignore


def format_columns(data: t.Any, prefix=""):
    if isinstance(data, t.Dict):
        return _format_columns_dict(data, prefix)

    raise NotImplementedError(format_columns, data, type(data))


def _format_columns_dict(dict, prefix):
    if not dict:
        return ""
    c1_width = max(len(k) for k in dict.keys())
    return "\n".join(
        "{}{:<{c1_width}}\t{}".format(prefix, k, v, c1_width=c1_width)
        for k, v in dict.items()
    ).strip("\n")


Color = t.NewType("Color", str)


class TERM_COLORS:
    NONE = Color("\033[0m")
    RESET = Color(NONE)

    RED = Color("\033[0;31m")
    ERROR = Color(RED)

    GREY = Color("\033[0;90m")
    DISABLED = Color(GREY)

    YELLOW = Color("\033[0;33m")
    WARNING = Color(YELLOW)


def colorize(string: str, color: Color):
    return color + string + TERM_COLORS.RESET



from __future__ import annotations  # ABCs are not generic prior to 3.9

import os
import typing as t
from collections.abc import MutableMapping
from configparser import ConfigParser
from contextlib import AbstractContextManager
from dataclasses import dataclass

import click

from mons.install import Install


config_dir = click.get_app_dir("mons", roaming=False)

CONFIG_FILE = "config.ini"
INSTALLS_FILE = "installs.ini"
CACHE_FILE = "cache.ini"

# Config, Installs, Cache
Config_DEFAULT = {}
Install.DEFAULTS = {
    "PreferredBranch": "stable",
}


def get_default_install():
    return os.environ.get("MONS_DEFAULT_INSTALL", None)


@dataclass
class Env:
    skip_confirmation = False
    ignore_errors = False


def loadConfig(file: str, default: MutableMapping[str, str] = {}):
    config = ConfigParser()
    config_file = os.path.join(config_dir, file)
    if os.path.isfile(config_file):
        config.read(config_file)
    else:
        config["DEFAULT"] = default
        os.makedirs(config_dir, exist_ok=True)
        with open(config_file, "x") as f:
            config.write(f)
    return config


def saveConfig(config: ConfigParser, file: str):
    with open(os.path.join(config_dir, file), "w") as f:
        config.write(f)


def editConfig(config: ConfigParser, file: str):
    saveConfig(config, file)
    click.edit(
        filename=os.path.join(config_dir, file),
        editor=config.get("user", "editor", fallback=None),
    )
    return loadConfig(file, config["DEFAULT"])


class UserInfo(AbstractContextManager):  # pyright: ignore[reportMissingTypeArgument]
    def __enter__(self):
        self.config = loadConfig(CONFIG_FILE, Config_DEFAULT)
        installs = loadConfig(INSTALLS_FILE)
        cache = loadConfig(CACHE_FILE)

        def load_install(name: str):
            if not cache.has_section(name):
                cache.add_section(name)
            return Install(
                name, installs[name]["Path"], cache=cache[name], data=installs[name]
            )

        self.installs = {name: load_install(name) for name in installs.sections()}
        if not self.config.has_section("user"):
            self.config["user"] = {}
        return self

    def __exit__(self, *exec_details):
        saveConfig(self.config, CONFIG_FILE)
        installs = ConfigParser()
        cache = ConfigParser()
        for k, v in self.installs.items():
            installs[k] = v.serialize()
            cache[k] = v.cache
        saveConfig(installs, INSTALLS_FILE)
        saveConfig(cache, CACHE_FILE)


pass_userinfo = click.make_pass_decorator(UserInfo)



import warnings

import click


def test_command_help(command: "click.Command"):
    if command.hidden:
        return

    assert command.help, f"Command '{command.name}' is missing help text"
    assert (
        command.get_short_help_str()
    ), f"Command '{command.name}' does is missing short help text"


def test_option_help(command: "click.Command"):
    if command.hidden:
        return

    for param in command.params:
        if isinstance(param, click.Option) and not param.hidden:
            assert (
                param.help
            ), f"Option '{param.name}' for command '{command.name}' is missing help text"


def test_no_args_is_help(command: "click.Command"):
    if command.hidden:
        return
    # Note: commands with only a clickExt.OptionalArg required do *not* fail this test. When an optional arg is applied 'no_args_is_help' is disabled.

    if command.no_args_is_help:
        assert any(
            param.required for param in command.params
        ), f"Command can be called with no arguments, but has 'no_args_is_help' enabled."
    elif any(param.required for param in command.params):
        warnings.warn(f"Command does not have 'no_args_is_help' enabled, but could.")



import os
import shutil
import typing as t
import urllib.request
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from contextlib import nullcontext
from io import BytesIO
from tempfile import TemporaryDirectory
from urllib.error import URLError

import urllib3
from click import Abort
from tqdm import tqdm
from urllib3.exceptions import HTTPError

import mons.fs as fs
from mons import baseUtils  # required to set module variable
from mons.baseUtils import read_with_progress
from mons.modmeta import ModDownload
from mons.modmeta import UpdateInfo


def get_download_size(
    url: str, initial_size=0, http_pool: t.Optional[urllib3.PoolManager] = None
):
    http = http_pool or urllib3.PoolManager()
    response: urllib3.HTTPResponse = http.request("HEAD", url)
    return int(response.headers.get("Content-Length")) - initial_size


def download_with_progress(
    src: t.Union[str, urllib.request.Request, urllib3.HTTPResponse],
    dest: t.Optional[str],
    label: t.Optional[str] = None,
    atomic: t.Optional[bool] = False,
    clear=False,
    *,
    response_handler: t.Optional[
        t.Callable[[urllib3.HTTPResponse], urllib3.HTTPResponse]
    ] = None,
    pool_manager: t.Optional[urllib3.PoolManager] = None,
):
    if not dest and atomic:
        raise ValueError("atomic download cannot be used without destination file")

    http = pool_manager or urllib3.PoolManager()

    if isinstance(src, (str, urllib.request.Request)):
        try:
            response: urllib3.HTTPResponse = http.request(
                "GET",
                src,
                preload_content=False,
                timeout=urllib3.Timeout(connect=3, read=10),
            )
        except HTTPError as e:
            try:
                # try to fallback to urlopen, notably for file:// scheme
                response = urllib.request.urlopen(src)
            except URLError:
                raise e
    else:
        response = src

    content = response_handler(response) if response_handler else response
    size = int(response.headers.get("Content-Length", None) or 100)
    blocksize = 8192

    with fs.temporary_file(persist=False) if atomic else nullcontext(dest) as file:
        with open(file, "wb") if file else nullcontext(BytesIO()) as io:
            read_with_progress(content, io, size, blocksize, label, clear)

            if dest is None and isinstance(io, BytesIO):
                # io will not be closed by contextmanager because it used nullcontext
                io.seek(0)
                return io

        if atomic:
            dest = t.cast(str, dest)
            if os.path.isfile(dest):
                os.remove(dest)
            shutil.move(t.cast(str, file), dest)

    return BytesIO()


def downloader(
    src: str,
    dest: str,
    name: str,
    mirror: t.Optional[str] = None,
    http_pool: t.Optional[urllib3.PoolManager] = None,
):
    mirror = mirror or src

    if os.path.isdir(dest):
        tqdm.write(f"\nCould not overwrite unzipped mod: {os.path.basename(dest)}")
    try:
        download_with_progress(
            src, dest, f"{name} {src}", atomic=True, clear=True, pool_manager=http_pool
        )
    except Abort:
        return
    except Exception as e:
        tqdm.write(f"\nError downloading file {os.path.basename(dest)} {src}: {e}")
        if isinstance(e, (HTTPError)) and src != mirror:
            downloader(mirror, dest, name)


def mod_downloader(
    mod_folder: str,
    download: t.Union[ModDownload, UpdateInfo],
    http_pool: urllib3.PoolManager,
):
    if isinstance(download, ModDownload):
        src, mirror, dest, name = (
            download.Url,
            download.Mirror,
            os.path.join(mod_folder, download.Meta.Name + ".zip"),
            str(download.Meta),
        )
    else:
        src, mirror, dest, name = (
            download.Url,
            download.Mirror,
            download.Old.Path,
            str(download.Old),
        )

    downloader(src, dest, name, mirror, http_pool)


def download_threaded(
    mod_folder: str,
    downloads: t.Sequence[t.Union[ModDownload, UpdateInfo]],
    late_downloads: t.Optional[t.Sequence[t.Union[ModDownload, UpdateInfo]]] = None,
    thread_count=8,
):
    http_pool = urllib3.PoolManager(maxsize=thread_count)
    with ThreadPoolExecutor(
        max_workers=thread_count, thread_name_prefix="download_"
    ) as pool:
        futures = [
            pool.submit(mod_downloader, mod_folder, download, http_pool)
            for download in downloads
        ]
        with TemporaryDirectory("_mons") if late_downloads else nullcontext(
            ""
        ) as temp_dir:
            if late_downloads:
                futures += [
                    pool.submit(mod_downloader, temp_dir, download, http_pool)
                    for download in late_downloads
                ]
            try:
                while True:
                    _, not_done = wait(futures, timeout=0.1)
                    if len(not_done) < 1:
                        break
            except (KeyboardInterrupt, SystemExit):
                for future in futures:
                    future.cancel()
                baseUtils._download_interrupt = (  # pyright: ignore[reportPrivateUsage]
                    True
                )
                raise

            if late_downloads:
                for file in os.listdir(temp_dir):
                    shutil.move(os.path.join(temp_dir, file), mod_folder)



import io
import os
import shutil
import subprocess
import typing as t
import urllib.parse
import zipfile

import click
from click import echo

import mons.clickExt as clickExt
import mons.fs as fs
from mons.config import config_dir
from mons.config import CONFIG_FILE
from mons.config import editConfig
from mons.config import pass_userinfo
from mons.config import UserInfo
from mons.downloading import download_with_progress
from mons.errors import TTYError
from mons.formatting import format_columns
from mons.install import Install
from mons.mons import cli
from mons.utils import build_exists
from mons.utils import fetch_build_artifact
from mons.utils import find_celeste_file
from mons.utils import getMD5Hash
from mons.utils import parseVersionSpec
from mons.utils import unpack


@cli.command(no_args_is_help=True)
@click.argument("name", type=clickExt.Install(exist=False))
@click.argument("path", type=click.Path(exists=True, resolve_path=True))
@pass_userinfo
def add(userInfo: UserInfo, name: str, path: str):
    """Add a Celeste install"""
    try:
        install_path = find_celeste_file(path, "Celeste.exe")
    except FileNotFoundError as err:
        raise click.UsageError(str(err))

    new_install = Install(name, install_path)
    userInfo.installs[name] = new_install

    echo(f"Found Celeste.exe: {install_path}")
    echo(new_install.version_string())


@cli.command(no_args_is_help=True)
@click.argument("name", type=clickExt.Install())
@click.option(
    "-e",
    "--eval",
    is_flag=True,
    help="Print `export MONS_DEFAULT_INSTALL={NAME}` to stdout.",
)
def use(name: str, eval: bool):
    """Set the default install for mons commands

    To un-set, run `export MONS_DEFAULT_INSTALL=`"""
    if eval:
        echo(f"export MONS_DEFAULT_INSTALL={name}")
    else:
        echo(
            f"""Mons can't set environment variables in the parent shell.
To circumvent this, run the following:

eval "$(mons use {name} --eval)" """,
            err=True,
        )


@cli.command(no_args_is_help=True)
@click.argument("old", type=clickExt.Install(exist=True))
@click.argument("new", type=clickExt.Install(exist=False))
@pass_userinfo
def rename(userInfo: UserInfo, old: str, new: str):
    """Rename a Celeste install"""
    userInfo.installs[new] = userInfo.installs.pop(old)
    userInfo.installs[new].name = new
    echo(f"Renamed install `{old}` to `{new}`")


@cli.command(no_args_is_help=True)
@click.argument("name", type=clickExt.Install(check_path=False, resolve_install=True))
@click.argument("path", type=click.Path(exists=True, resolve_path=True))
def set_path(name: Install, path: str):
    """Change the path of an existing install"""
    try:
        install_path = find_celeste_file(path, "Celeste.exe")
    except FileNotFoundError as err:
        raise click.UsageError(str(err))

    name.path = install_path
    echo(f"Found Celeste.exe: {install_path}")
    echo(name.version_string())


@cli.command(
    no_args_is_help=True,
)
@click.argument("name", type=clickExt.Install(check_path=False))
@click.confirmation_option(
    "--force",
    prompt="Are you sure you want to remove this install?",
    help="Skip confirmation prompt.",
)
@pass_userinfo
def remove(userInfo: UserInfo, name: str):
    """Remove an existing install"""
    del userInfo.installs[name]
    echo(f"Removed install {name}.")


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
@click.argument("branch")
def set_branch(name: Install, branch: str):
    """Set the preferred branch name for an existing install"""
    name["PreferredBranch"] = branch
    echo(f"Preferred branch for `{name.name}` is now `{branch}`.")


@cli.command()
# @click.option("--json", is_flag=True, hidden=True)
@pass_userinfo
def list(userInfo: UserInfo):
    """List existing installs"""
    output = {}
    if not userInfo.installs:
        raise click.ClickException("No installs found, use `add` to add one.")

    for name, install in userInfo.installs.items():
        try:
            clickExt.Install.validate_install(name, validate_path=True)
            output[name] = install.version_string()
        except Exception as err:
            raise click.UsageError(str(err))

    click.echo(format_columns(output))


@cli.command(cls=clickExt.CommandExt, no_args_is_help=True)
@clickExt.install("name")
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging.")
def show(name: Install, verbose: bool):
    """Display information for a specific install"""
    install = name
    if verbose:
        echo(install.name + ":")
        output = {**install.get_cache()}
        orig_exe = os.path.join(os.path.dirname(install.path), "orig", "Celeste.exe")
        if os.path.isfile(orig_exe):
            output["vanilla hash"] = getMD5Hash(orig_exe)
        output["path"] = install.path
        echo(format_columns(output, prefix="\t"))
    else:
        echo("{}:\t{}".format(install.name, install.version_string()))
        echo(install.path)


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
@click.argument("versionSpec", required=False)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging.")
@click.option(
    "--latest", is_flag=True, help="Install latest available build, branch-ignorant."
)
@click.option("--zip", type=click.File(mode="rb"), help="Install from zip artifact.")
@click.option(
    "--url",
    type=clickExt.URL(require_path=True),
    help="Download and install from a URL.",
)
@click.option(
    "--src",
    cls=clickExt.ExplicitOption,
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Build and install from source folder.",
)
@click.option("--src", cls=clickExt.DefaultOption, is_flag=True)
@click.option(
    "--no-build", is_flag=True, help="Use with --src to install without building."
)
@click.option("--launch", is_flag=True, help="Launch Celeste after installing.")
@pass_userinfo
def install(
    userinfo: UserInfo,
    name: Install,
    versionspec: str,
    verbose: bool,
    latest: bool,
    zip: io.BufferedReader,
    url: t.Optional[urllib.parse.ParseResult],
    src: t.Optional[str],
    src_default: bool,
    no_build: bool,
    launch: bool,
):
    """Install Everest

    VERSIONSPEC can be a branch name, build number, or version number."""
    install = name
    path = install.path
    installDir = os.path.dirname(path)
    success = False

    artifactPath = None
    build = None

    if src_default:
        src = userinfo.config.get("user", "SourceDirectory", fallback=None)
        if not src:
            raise click.BadOptionUsage(
                "--src", "--src option passed with no path and no SourceDirectory set"
            )

    if src:
        build_success = 0 if no_build else 1
        if not no_build:
            if shutil.which("dotnet"):
                build_success = subprocess.run(
                    [
                        "dotnet",
                        "build",
                        "--verbosity",
                        "normal" if verbose else "minimal",
                    ],
                    cwd=src,
                ).returncode
            elif shutil.which("msbuild"):
                build_success = subprocess.run(
                    ["msbuild", "-verbosity:" + ("normal" if verbose else "minimal")],
                    cwd=src,
                ).returncode
            else:
                raise click.ClickException(
                    "Unable to build project: could not find `dotnet` or `msbuild` on PATH.\n"
                    + "Include the --no-build switch to skip build step."
                )

        if build_success == 0:
            echo("Copying files...")
            fs.copy_recursive_force(
                os.path.join(src, "Celeste.Mod.mm", "bin", "Debug", "net452"),
                installDir,
                ignore=lambda path, names: [
                    name
                    for name in names
                    if fs.is_unchanged(
                        os.path.join(path, name), os.path.join(installDir, name)
                    )
                ],
            )
            fs.copy_recursive_force(
                os.path.join(src, "MiniInstaller", "bin", "Debug", "net452"),
                installDir,
                ignore=lambda path, names: [
                    name
                    for name in names
                    if fs.is_unchanged(
                        os.path.join(path, name), os.path.join(installDir, name)
                    )
                ],
            )
            success = True

    elif url:
        download_url = urllib.parse.urlunparse(url)
        echo("Downloading artifact from " + download_url)
        artifactPath = os.path.join(installDir, url.path.split("/")[-1])
        download_with_progress(download_url, artifactPath, atomic=True, clear=True)

    elif zip:
        artifactPath = zip

    if artifactPath:
        label = f"Unzipping {os.path.basename(artifactPath) if isinstance(artifactPath, str) else zip.name}"
        if zip and zip.fileno() == 0:  # stdin
            if zip.isatty():
                raise TTYError("no input.")
            artifactPath = io.BytesIO(zip.read())
        with zipfile.ZipFile(artifactPath) as wrapper:
            try:
                entry = wrapper.open(
                    "olympus-build/build.zip"
                )  # Throws KeyError if not present
                with zipfile.ZipFile(entry) as artifact:
                    unpack(artifact, installDir, label=label)
                    success = True
            except KeyError:
                unpack(wrapper, installDir, "main/", label=label)
                success = True

    elif not src:
        versionspec = "" if latest else (versionspec or install["PreferredBranch"])
        build = parseVersionSpec(versionspec)
        if not build:
            raise click.ClickException(
                f"Build number could not be retrieved for `{versionspec}`."
            )

        if str(build) == install.cache.get("EverestBuild"):
            echo(f"Build {build} already installed.")
            exit(0)

        if not build_exists(build):
            raise click.ClickException(
                f"Build artifacts could not be found for build {build}."
            )

        echo(f"Installing Everest build {build}")
        echo("Downloading build metadata...")
        try:
            meta = fetch_build_artifact(build, "olympus-meta")
            with zipfile.ZipFile(io.BytesIO(meta.read())) as file:
                size = int(file.read("olympus-meta/size.txt"))
        except:
            size = 0

        if size > 0:
            echo("Downloading olympus-build.zip", nl=False)
            response = fetch_build_artifact(build, "olympus-build")
            response.headers["Content-Length"] = str(size)
            artifactPath = os.path.join(installDir, "olympus-build.zip")
            echo(f" to file {artifactPath}")
            download_with_progress(
                response, artifactPath, label="Downloading", clear=True
            )
            with zipfile.ZipFile(artifactPath) as wrapper:
                with zipfile.ZipFile(
                    wrapper.open("olympus-build/build.zip")
                ) as artifact:
                    unpack(artifact, installDir, label="Extracting")
                    success = True

        else:
            echo("Downloading main.zip", nl=False)
            response = fetch_build_artifact(build, "main")
            artifactPath = os.path.join(installDir, "main.zip")
            echo(f" to file {artifactPath}")
            download_with_progress(
                response, artifactPath, label="Downloading", clear=True
            )
            echo("Unzipping main.zip")
            with zipfile.ZipFile(artifactPath) as artifact:
                unpack(artifact, installDir, "main/", label="Extracting")
                success = True

    if success:
        echo("Running MiniInstaller...")
        stdout = None if verbose else subprocess.DEVNULL
        if os.name == "nt":
            installer_ret = subprocess.run(
                os.path.join(installDir, "MiniInstaller.exe"),
                stdout=stdout,
                stderr=None,
                cwd=installDir,
            )
        else:
            uname = os.uname()
            if uname.sysname == "Darwin":
                kickstart_dir = os.path.join(installDir, "..", "MacOS")
                with fs.copied_file(
                    os.path.join(kickstart_dir, "Celeste"),
                    os.path.join(kickstart_dir, "MiniInstaller"),
                ) as miniinstaller:
                    installer_ret = subprocess.run(
                        miniinstaller, stdout=stdout, stderr=None, cwd=installDir
                    )
            else:
                suffix = "x86_64" if uname.machine == "x86_64" else "x86"
                with fs.copied_file(
                    os.path.join(os.path.join(installDir, f"Celeste.bin.{suffix}")),
                    os.path.join(installDir, f"MiniInstaller.bin.{suffix}"),
                ) as miniinstaller:
                    installer_ret = subprocess.run(
                        miniinstaller, stdout=stdout, stderr=None, cwd=installDir
                    )

        if installer_ret.returncode == 0:
            echo("Install success")
            if build:
                peHash = getMD5Hash(path)
                install.cache.update(
                    {
                        "Hash": peHash,
                        "Everest": str(True),
                        "EverestBuild": str(build),
                    }
                )
            else:
                install.update_cache()
                echo("Install info cached")
            if launch:
                echo("Launching Celeste...")
                cli.main(args=["launch", install.name])
            return

    # If we got this far, something went wrong
    click.get_current_context().exit(1)


@cli.command(
    no_args_is_help=True,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
    cls=clickExt.CommandExt,
)
@clickExt.install("name")
@click.argument("args", nargs=-1, required=False, cls=clickExt.PlaceHolder)
@click.pass_context
def launch(ctx: click.Context, name: Install):
    """Launch the game associated with an install

    Any additional arguments are passed to the launched process."""
    path = name.path
    if os.name != "nt":
        if os.uname().sysname == "Darwin":
            path = os.path.normpath(
                os.path.join(os.path.dirname(path), "..", "MacOS", "Celeste")
            )
        else:
            path = os.path.splitext(path)[0]  # drop the .exe

    redirect = None if "--console" in ctx.args else subprocess.PIPE

    proc = subprocess.Popen(
        [path] + ctx.args, stdout=redirect, stderr=redirect, shell=True
    )
    if not redirect:
        proc.wait()


@cli.command()
@click.option(
    "-e", "--edit", is_flag=True, help="Open the global config file for editing."
)
@click.option("--open", is_flag=True, help="Show the mons config folder.")
@pass_userinfo
def config(userinfo: UserInfo, edit: bool, open: bool):
    """Manage the global config"""
    if edit:
        userinfo.config = editConfig(userinfo.config, CONFIG_FILE)
    elif open:
        click.launch(os.path.join(config_dir, CONFIG_FILE), locate=True)
    else:
        raise click.UsageError(
            """Managing config directly via commandline is not currently supported.
Use --edit to edit the config using the default editor."""
        )



import click


class EmptyFileError(Exception):
    pass


class TTYError(click.ClickException):
    def __init__(self, message: str) -> None:
        super().__init__("Could not read from stdin: " + message)



from dataclasses import dataclass


@dataclass(frozen=True)
class Version:
    Major: int
    Minor: int
    Build: int = -1
    Revision: int = -1

    @classmethod
    def parse(cls, version: str):
        if version == "NoVersion":
            return cls(1, 0)

        # discard semver prerelease version
        strArr = version.split("-", maxsplit=1)[0].split(".")
        if 4 > len(strArr) < 2 or not all(n.isdigit() for n in strArr):
            raise ValueError("%s is not a valid Version string" % version)
        arr = list(map(int, strArr))
        arr += [-1] * (4 - len(arr))
        return cls(*arr)

    def satisfies(self, required: "Version"):
        """Checks if this version satisfies the :param:`required` version."""
        # Special case: Always True if version == 0.0.*
        if self.Major == 0 and self.Minor == 0:
            return True

        # Major version, breaking changes, must match.
        if self.Major != required.Major:
            return False
        # Minor version, non-breaking changes, installed can't be lower than what we depend on.
        if self.Minor < required.Minor:
            return False
        # "Build" is "PATCH" in semver, but we'll also check for it and "Revision".
        if self.Minor == required.Minor and self.Build < required.Build:
            return False
        if (
            self.Minor == required.Minor
            and self.Build == required.Build
            and self.Revision < required.Revision
        ):
            return False

        return True

    def __str__(self):
        out = "{}.{}".format(self.Major, self.Minor)
        if self.Build != -1:
            out += ".{}".format(self.Build)
            if self.Revision != -1:
                out += ".{}".format(self.Revision)
        return out

    def __gt__(self, other: "Version"):
        if self.Major > other.Major:
            return True
        elif self.Major == other.Major and self.Minor > other.Minor:
            return True
        elif (
            self.Major == other.Major
            and self.Minor == other.Minor
            and self.Build > other.Build
        ):
            return True
        elif (
            self.Major == other.Major
            and self.Minor == other.Minor
            and self.Build == other.Build
            and self.Revision > other.Revision
        ):
            return True
        return False



import os
import typing as t
import zipfile

import xxhash
import yaml
from click import echo
from yaml.scanner import ScannerError

import mons.fs as fs
from mons.baseUtils import find
from mons.errors import EmptyFileError
from mons.version import Version


class _ModMeta_Base:
    def __init__(self, name: str, version: t.Union[str, Version]):
        self.Name = name
        if isinstance(version, str):
            version = Version.parse(version)
        self.Version = version

    @classmethod
    def _from_dict(cls, data: t.Dict[str, t.Any]):
        return cls(str(data["Name"]), str(data.get("Version", "NoVersion")))

    def __repr__(self) -> str:
        return f"{self.Name}: {self.Version}"


class _ModMeta_Deps:
    def __init__(
        self,
        dependencies: t.List[_ModMeta_Base],
        optionals: t.List[_ModMeta_Base],
    ):
        self.Dependencies = dependencies
        self.OptionalDependencies = optionals
        assert isinstance(self.Dependencies, t.List)

    @classmethod
    def parse(cls, data: t.Any):
        if isinstance(data, cls):
            return data
        elif isinstance(data, t.Dict):
            return cls._from_dict(data)
        elif isinstance(data, t.List):
            return cls(data, [])
        else:
            raise ValueError()

    @classmethod
    def _from_dict(cls, data: t.Dict[str, t.Any]):
        return cls(
            [
                _ModMeta_Base._from_dict(dep)  # pyright: ignore[reportPrivateUsage]
                for dep in data["Dependencies"]
            ],
            [
                _ModMeta_Base._from_dict(dep)  # pyright: ignore[reportPrivateUsage]
                for dep in data["OptionalDependencies"]
            ],
        )


class ModMeta(_ModMeta_Base, _ModMeta_Deps):
    Hash: t.Optional[str]
    Path: str
    Blacklisted: t.Optional[bool] = False

    def __init__(self, data: t.Dict[str, t.Any]):
        _ModMeta_Base.__init__(
            self, str(data["Name"]), str(data.get("Version", "NoVersion"))
        )
        _ModMeta_Deps.__init__(
            self,
            [_ModMeta_Base._from_dict(dep) for dep in data.get("Dependencies", [])],
            [
                _ModMeta_Base._from_dict(dep)
                for dep in data.get("OptionalDependencies", [])
            ],
        )
        self.DLL = str(data["DLL"]) if "DLL" in data else None
        self.Size = int(data["Size"]) if "Size" in data else 0

    @classmethod
    def placeholder(cls, path: str):
        basename = os.path.basename(path)
        meta = None
        if os.path.isdir(path):
            meta = cls({"Name": "_dir_" + basename, "Version": Version(0, 0, 0)})

        elif zipfile.is_zipfile(path):
            name = os.path.splitext(basename)[0]
            meta = cls({"Name": "_zip_" + name, "Version": Version(0, 0, 0)})

        if meta:
            meta.Path = path
            return meta

        return None


class ModDownload:
    def __init__(
        self,
        meta: t.Union[ModMeta, t.Dict[str, t.Any]],
        url: str,
        mirror: t.Optional[str] = None,
    ):
        if isinstance(meta, t.Dict):
            meta = ModMeta(meta)
        self.Meta = meta
        self.Url = url
        self.Mirror = mirror if mirror else url


def _merge_dependencies(dict: t.Dict[str, _ModMeta_Base], dep: _ModMeta_Base):
    if dep.Name in dict:
        if dep.Version.Major != dict[dep.Name].Version.Major:
            raise ValueError(
                "Incompatible dependencies encountered: "
                + f"{dep.Name} {dep.Version} vs {dep.Name} {dict[dep.Name].Version}"
            )
        elif dep.Version > dict[dep.Name].Version:
            dict[dep.Name] = dep
    else:
        dict[dep.Name] = dep


def recurse_dependencies(
    mods: t.Iterable[_ModMeta_Base],
    database: t.Dict[str, t.Any],
    dict: t.Dict[str, ModMeta],
):
    for mod in mods:
        _merge_dependencies(dict, mod)  # type: ignore
        if mod.Name in database:
            recurse_dependencies(
                _ModMeta_Deps.parse(database[mod.Name]).Dependencies, database, dict
            )


def combined_dependencies(
    mods: t.Iterable[_ModMeta_Base], database: t.Dict[str, t.Any]
) -> t.Dict[str, _ModMeta_Base]:
    deps = {}
    for mod in mods:
        dependencies = None
        if mod.Name in database:
            dependencies = _ModMeta_Deps.parse(database[mod.Name]).Dependencies
        elif isinstance(mod, _ModMeta_Deps):
            dependencies = mod.Dependencies
        if dependencies:
            recurse_dependencies(dependencies, database, deps)
    return deps


class UpdateInfo:
    def __init__(
        self, old: ModMeta, new: Version, url: str, mirror: t.Optional[str] = None
    ):
        self.Old = old
        self.New = new
        self.Url = url
        self.Mirror = mirror if mirror else url


def read_mod_info(mod: t.Union[str, t.IO[bytes]], with_size=False, with_hash=False):
    meta = None
    try:
        if not isinstance(mod, str) or os.path.isfile(mod) and zipfile.is_zipfile(mod):
            with zipfile.ZipFile(mod) as zip:
                everest_file = find(zip.namelist(), ("everest.yaml", "everest.yml"))
                if everest_file:
                    yml = yaml.safe_load(zip.read(everest_file).decode("utf-8-sig"))
                    if yml is None:
                        raise EmptyFileError()
                    meta = ModMeta(yml[0])
                    if zip.fp:
                        zip.fp.seek(0)
                        if with_hash:
                            meta.Hash = xxhash.xxh64_hexdigest(zip.fp.read())
                        zip.fp.seek(0, os.SEEK_END)
                        meta.Size = zip.fp.tell() if with_size else 0

        elif os.path.isdir(mod):
            everest_file = fs.find_file(mod, ("everest.yaml", "everest.yml"))
            if everest_file:
                with open(
                    os.path.join(mod, everest_file), encoding="utf-8-sig"
                ) as file:
                    yml = yaml.safe_load(file)
                    if yml is None:
                        raise EmptyFileError()
                    meta = ModMeta(yml[0])
                meta.Size = fs.folder_size(mod) if with_size else 0
    except (EmptyFileError, ScannerError):
        return None
    except Exception:
        echo(mod)
        raise

    if meta:
        meta.Path = mod if isinstance(mod, str) else ""
    return meta



import os
import sys
import typing as t
from traceback import format_exception_only
from traceback import format_tb
from urllib import parse

import click

from mons.baseUtils import *
from mons.config import Env
from mons.config import get_default_install
from mons.config import UserInfo
from mons.errors import TTYError
from mons.formatting import colorize
from mons.formatting import TERM_COLORS
from mons.install import Install as T_Install


def confirm_ext(*params, default, dangerous: bool = False, **attrs):
    """Extension to :func:`click.confirm`.

    Throws a :class:`TTYError` if `stdin` is not a TTY,
    and returns `True` if :attr:`Env.ignore_errors` is set.

    :param dangerous: if set to `False` (default), :attr:`Env.skip_confirmation`
    will also be checked.
    """

    ctx = click.get_current_context(silent=True)
    env = ctx and ctx.find_object(Env)

    if env:
        if env.ignore_errors:
            return True
        elif env.skip_confirmation and not dangerous:
            return True

    if not os.isatty(sys.stdin.fileno()):
        if dangerous:
            msg = "Use '--force' to skip error prompts."
        else:
            msg = "Use '--yes' to skip confirmation prompts."
        raise TTYError("not a tty.\n" + msg)

    return click.confirm(default=default, *params, **attrs)


def env_flag_option(
    var: str, *param_decls: str, help="", process_value: t.Any = None, **kwargs: t.Any
):
    def callback(ctx: click.Context, param: click.Parameter, value: bool):
        env = ctx.find_object(Env)
        if process_value:
            value = process_value(ctx, param, value)
        setattr(env, var, value)

    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", help)
    kwargs["callback"] = callback
    return click.option(*param_decls, **kwargs)


def yes_option(*param_decls: str, **kwargs: t.Any):
    if not param_decls:
        param_decls = ("--yes",)

    kwargs.setdefault("is_flag", True)
    return env_flag_option(
        "skip_confirmation", *param_decls, help="Skip confirmation prompts.", **kwargs
    )


def force_option(*param_decls: str, **kwargs: t.Any):
    if not param_decls:
        param_decls = ("--force",)

    kwargs.setdefault("is_flag", True)
    return env_flag_option(
        "ignore_errors",
        *param_decls,
        help="Ignore errors and confirmation prompts.",
        **kwargs,
    )


class CatchErrorsGroup(click.Group):
    def main(self, args=None, *params, **extra):
        def pop_arg(arg):
            if arg in sys.argv:
                if not args:
                    sys.argv.remove(arg)
                return True
            return False

        debug = pop_arg("--debug")
        wait = pop_arg("--wait")
        if pop_arg("--prompt-install"):
            os.environ["MONS_PROMPT_INSTALL"] = "1"
        try:
            super().main(args=args, *params, **extra)
        except SystemExit as e:
            if wait:
                click.pause()
            raise e
        except Exception as e:
            if debug or os.environ.get("MONS_DEBUG", "false") == "true":
                click.echo(
                    colorize("An unhandled exception has occurred.", TERM_COLORS.ERROR)
                )
                click.echo("".join(format_tb(e.__traceback__)), nl=False)
                click.echo(
                    colorize(
                        "".join(format_exception_only(type(e), e)), TERM_COLORS.ERROR
                    ),
                    nl=False,
                )
            else:
                click.echo(colorize(repr(e), TERM_COLORS.ERROR))
                click.echo(
                    f"""An unhandled exception has occurred.
Use the --debug flag to disable clean exception handling."""
                )


def color_option(*param_decls: str, **kwargs: t.Any):
    def auto_color():
        if os.environ.get("NO_COLOR"):
            return False
        # TODO: Special cases for pagers not covered by click

        # Default to whatever click decides
        return None

    def callback(ctx, param, value: t.Optional[str]):
        if value is None:
            return auto_color()

        if value == "always":
            return True
        if value == "never":
            return False
        if value == "auto":
            return None
        raise click.BadParameter("Possible values: auto, never, always", ctx, param)

    if not param_decls:
        param_decls = ("--color",)

    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", "Specify when to use colored output: auto, always, none")
    kwargs.setdefault("metavar", "WHEN")
    kwargs["callback"] = callback
    return click.option(*param_decls, **kwargs)


class Install(click.ParamType):
    name = "Install"

    def __init__(self, exist=True, resolve_install=False, check_path=True) -> None:
        super().__init__()
        self.exist = exist
        self.resolve_install = resolve_install
        self.validate_path = check_path

    def convert(self, value: t.Union[str, T_Install], param, ctx):
        userinfo = ctx and ctx.find_object(UserInfo)
        if not userinfo:
            return value
        installs = userinfo.installs

        if self.exist:
            if not isinstance(value, T_Install):
                try:
                    Install.validate_install(
                        value, validate_path=self.validate_path, ctx=ctx
                    )
                except ValueError as err:
                    self.fail(str(err), param, ctx)
                except FileNotFoundError as err:
                    raise click.UsageError(str(err), ctx)

                if self.resolve_install:
                    value = installs[value]
        else:
            if value in installs:
                self.fail(f"Install {value} already exists.", param, ctx)

        return value

    @classmethod
    def validate_install(cls, install: str, validate_path=True, ctx=None):
        userinfo: UserInfo = (ctx or click.get_current_context()).obj
        installs = userinfo.installs

        if install not in installs:
            raise ValueError(f"Install {install} does not exist")

        path = installs[install].path
        if validate_path:
            error = None
            if not os.path.exists(path):
                error = "does not exist."
            elif not os.path.basename(path) == "Celeste.exe":
                error = "does not point to Celeste.exe"

            if error:
                raise FileNotFoundError(
                    f"""Install {install} does not have a valid path:
{TERM_COLORS.ERROR}{path} {error}{TERM_COLORS.RESET}
Use `set-path` to assign a new path."""
                )


def install(*param_decls, resolve=True, **attrs):
    """Alias for a `click.argument` of type `Install` that will use the default provided by `MONS_DEFAULT_INSTALL`

    Requires a command `cls` of `CommandExt`."""
    return click.argument(
        *param_decls,
        type=Install(resolve_install=resolve),
        cls=OptionalArg,
        default=get_default_install,
        prompt="Install name",
        warning="mons default install set to {default}",
        prompt_envvar="MONS_PROMPT_INSTALL",
        **attrs,
    )


class URL(click.ParamType):
    name = "URL"

    def __init__(
        self,
        default_scheme: t.Optional[str] = None,
        valid_schemes: t.Optional[t.Collection[str]] = None,
        require_path=False,
    ) -> None:
        super().__init__()
        self.default_scheme = default_scheme
        self.valid_schemes = valid_schemes
        self.require_path = require_path

    def convert(self, value: t.Union[str, parse.ParseResult], param, ctx):
        if isinstance(value, parse.ParseResult):
            return value

        try:
            parsed_url = parse.urlparse(value)

            if self.require_path and not parsed_url.path:
                self.fail("Path component required for URL.", param, ctx)
            if not parsed_url.scheme and self.default_scheme:
                if not value.startswith("//"):
                    # urlparse treats urls NOT starting with // as relative URLs
                    # https://docs.python.org/3.10/library/urllib.parse.html?highlight=urlparse#urllib.parse.urlparse
                    parsed_url = parse.urlparse(
                        "//" + value, scheme=self.default_scheme
                    )
                else:
                    parsed_url._replace(scheme=self.default_scheme)
            if self.valid_schemes and parsed_url.scheme not in self.valid_schemes:
                self.fail(f"URI scheme '{parsed_url.scheme}' not allowed.", param, ctx)

            return parsed_url
        except ValueError:
            self.fail(f"{value} is not a valid URL.", param, ctx)


class DefaultOption(click.Option):
    """Mark this option as being a _default option"""

    register_default = True

    def __init__(self, param_decls: t.Sequence[str], **attrs):
        param_decls = [decl + "_default" for decl in param_decls]
        super(DefaultOption, self).__init__(param_decls, **attrs)
        self.hidden = True


class ExplicitOption(click.Option):
    """Fix the help string for this option to display as an optional argument"""

    def get_help_record(self, ctx):
        help = super(ExplicitOption, self).get_help_record(ctx)
        if help:
            return (help[0].replace(" ", "[=", 1) + "]",) + help[1:]

    def sphinx_get_help_record(
        self, help_func: t.Callable[[click.Parameter], t.Tuple[str]]
    ):
        help = help_func(self)
        return (help[0].replace(" ", "[=", 1) + "]",) + help[1:]


class PlaceHolder(click.Argument):
    """Mark this argument as a placeholder that isn't processed"""

    register_placeholder = True


class OptionalArg(click.Argument):
    def __init__(
        self,
        param_decls: t.Sequence[str],
        **attrs,
    ) -> None:
        self.warning: t.Optional[str] = attrs.pop("warning", None)
        self.prompt: str = attrs.pop("prompt", "")
        self.prompt_envvar: str = attrs.pop("prompt_envvar", "")
        super().__init__(param_decls, True, **attrs)

    def should_prompt(self, ctx: click.Context):
        return os.environ.get(self.prompt_envvar, None) and not ctx.resilient_parsing

    def add_to_parser(self, parser: click.parser.OptionParser, ctx: click.Context):
        if not self.should_prompt(ctx):
            return super().add_to_parser(parser, ctx)

    def consume_value(self, ctx: click.Context, opts: t.Mapping[str, t.Any]):
        if self.should_prompt(ctx):
            source = click.core.ParameterSource.PROMPT
            default = self.get_default(ctx)
            value = click.prompt(
                self.prompt,
                default=default,
                type=self.type,
                value_proc=lambda x: self.process_value(ctx, x),
            )
            return value, source

        return super().consume_value(ctx, opts)


class CommandExt(click.Command):
    """Command implementation for extended option and argument types"""

    warnings: t.List[str] = list()

    def make_parser(self, ctx):
        """Strip placeholder params"""
        self.params = [
            a for a in self.params if not getattr(a, "register_placeholder", None)
        ]
        return super().make_parser(ctx)

    def parse_args(self, ctx, args):
        # Handle any OptionalArgs as needed
        def handle_optionalarg(o):
            if isinstance(o, OptionalArg) and o.default:
                default = (
                    o.default() if isinstance(o.default, t.Callable) else o.default
                )
                if default:
                    assert o.name
                    # set value for param directly in the context
                    ctx.params.update({o.name: o.type_cast_value(ctx, default)})
                    if o.warning:
                        self.warnings.append(o.warning.format(default=default))

                    # ensure that the command is invoked even if there are technically no arguments passed
                    self.no_args_is_help = False
                    return True
            return False

        ctx.command.params = [
            o for o in ctx.command.params if not handle_optionalarg(o)
        ]

        # Translate any opt to opt_default as needed
        options = [
            o for o in ctx.command.params if getattr(o, "register_default", None)
        ]
        prefixes = {
            p.replace("_default", "")
            for p in sum((o.opts for o in options), [])
            if p.startswith("--")
        }
        for i, a in enumerate(args):
            a = a.split("=", 1)
            if a[0] in prefixes and len(a) == 1:
                args[i] = a[0] + "_default"

        return super(CommandExt, self).parse_args(ctx, args)

    def invoke(self, ctx):
        """Emit additional warnings as needed"""
        for w in self.warnings:
            click.echo(colorize(w, TERM_COLORS.WARNING))
        return super().invoke(ctx)



import errno
import os
import pathlib

import pytest


def pytest_collection_modifyitems(session, config, items):
    module = pathlib.Path(os.path.dirname(__file__))
    for item in items:
        if module == item.path.parent:
            item.add_marker(pytest.mark.slow_integration_test)


# We only need to log this warning once for each test run
log_test_install_warning = True


@pytest.fixture(scope="function")
def test_install(pytestconfig: pytest.Config, cache: pytest.Cache):
    # prioritize cmdline arg so that it can be changed
    install_path = pytestconfig.getoption("mons_test_install", None)  # type: ignore
    if not install_path:
        install_path = cache.get("mons/test_install", None)

    if not install_path:
        global log_test_install_warning
        if log_test_install_warning:
            log_test_install_warning = True
            pytest.skip(
                "Test install not configured. Use the `--mons-test-install` cmdline flag to provide a test install."
            )
        pytest.skip()

    if os.path.exists(install_path):  # type: ignore
        cache.set("mons/test_install", install_path)
        return install_path
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), install_path)


@pytest.fixture(autouse=True)
def set_env(monkeypatch):
    monkeypatch.setenv("MONS_DEBUG", "true")


@pytest.fixture(autouse=True)
def isolated_filesystem(monkeypatch, tmp_path):
    """Isolate all referenced files to temp folder"""
    monkeypatch.setattr("mons.config.config_dir", os.path.join(tmp_path, ".config"))
    return {"config_dir": os.path.join(tmp_path, ".config")}



import itertools
import json
import os
import re
import shutil
import typing as t
import urllib.parse
import urllib.request
from gettext import ngettext

import click
from click import echo

import mons.clickExt as clickExt
import mons.fs as fs
from mons.baseUtils import *
from mons.downloading import download_threaded
from mons.downloading import download_with_progress
from mons.downloading import get_download_size
from mons.errors import TTYError
from mons.formatting import colorize
from mons.formatting import format_bytes
from mons.formatting import format_columns
from mons.formatting import TERM_COLORS
from mons.install import Install
from mons.modmeta import combined_dependencies
from mons.modmeta import ModDownload
from mons.modmeta import ModMeta
from mons.modmeta import read_mod_info
from mons.modmeta import UpdateInfo
from mons.mons import cli as mons_cli
from mons.utils import enable_mods
from mons.utils import get_dependency_graph
from mons.utils import get_mod_list
from mons.utils import installed_mods
from mons.utils import read_blacklist
from mons.utils import search_mods
from mons.utils import timed_progress
from mons.version import Version


@click.group(name="mods")
@click.pass_context
def cli(ctx: click.Context):
    """Manage Everest Mods
    \f

    |full_reference-mods|"""
    pass


mons_cli.add_command(cli)


def format_mod_info(meta: ModMeta):
    out = (
        colorize(f"{meta.Name}\t(disabled)", TERM_COLORS.DISABLED)
        if meta.Blacklisted
        else meta.Name
    ) + "\n"
    out += format_columns(
        {
            "Version": meta.Version,
            "File": os.path.basename(meta.Path)
            + ("/" if os.path.isdir(meta.Path) else ""),
        },
        prefix="\t",
    )
    out += "\n"
    return out


@cli.command(
    name="list",
    help="List installed mods.",
    no_args_is_help=True,
    cls=clickExt.CommandExt,
)
@clickExt.install("name")
@click.option(
    "--enabled/--disabled", help="Filter by enabled/disabled mods.", default=None
)
@click.option(
    "--valid/--invalid", help="Filter mods with valid everest.yaml.", default=None
)
@click.option("--dll/--no-dll", help="Filter mods that register DLLs.", default=None)
@click.option(
    "--dir/--zip",
    "dir",
    flag_value=True,
    help="Filter mods in folders/zips.",
    default=None,
)
@click.option("--no-zip/--no-dir", "dir", flag_value=False, hidden=True, default=None)
@click.option("-d", "--dependency", help="Filter mods by dependency.", metavar="MODID")
@click.option(
    "-s", "--search", help="Filter mods with a regex pattern.", metavar="QUERY"
)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging.")
@clickExt.color_option()
def list_mods(
    enabled: t.Optional[bool],
    valid: t.Optional[bool],
    name: Install,
    dll: t.Optional[bool],
    dir: t.Optional[bool],
    dependency: str,
    search: str,
    verbose: bool,
    color: t.Optional[bool],
):
    """List installed mods."""
    if valid == False:
        if dll == True:
            raise click.BadOptionUsage(
                "--dll", "--dll cannot be used with the --invalid flag."
            )
        if dependency:
            raise click.BadOptionUsage(
                "--dependency", "--dependency cannot be used with the --invalid flag."
            )

    basePath = os.path.join(os.path.dirname(name.path), "Mods")

    installed = installed_mods(
        basePath, dirs=dir, valid=valid, blacklisted=invert(enabled)
    )

    gen = installed
    if dll is not None:
        gen = filter(lambda meta: not dll ^ bool(meta.DLL), gen)

    if dependency:
        gen = filter(
            lambda meta: next(
                filter(lambda d: dependency == d.Name, meta.Dependencies),
                t.cast(t.Any, None),
            ),
            gen,
        )

    if search:
        pattern = re.compile(search, flags=re.I)
        gen = filter(
            lambda meta: bool(
                pattern.search(meta.Name) or pattern.search(os.path.basename(meta.Path))
            ),
            gen,
        )

    if verbose:
        gen = (format_mod_info(meta) for meta in gen)
    else:
        gen = (
            (
                f"{meta.Name}\t{meta.Version}"
                if not meta.Blacklisted
                else colorize(f"{meta.Name}\t(disabled)", TERM_COLORS.DISABLED)
            )
            + "\n"
            for meta in gen
        )

    click.echo_via_pager(gen, color=color)


@cli.command(hidden=True)
@click.argument("search")
def search(search: str):
    mod_list = get_mod_list()
    if search in mod_list:
        echo(mod_list[search]["GameBananaId"])
        return

    search_result = search_mods(search)
    for item in search_result:
        match = [
            mod
            for mod, data in mod_list.items()
            if data["GameBananaId"] == item["itemid"]
        ]
        for m in match:
            echo(m)
        if len(match) < 1:
            raise click.UsageError("Entry not found: " + str(item["itemid"]))


def prompt_mod_selection(options: t.Dict[str, t.Any], max=-1):
    matchKeys = sorted(
        options.keys(), key=lambda key: options[key]["LastUpdate"], reverse=True
    )
    selection = None
    if len(matchKeys) == 1:
        key = matchKeys[0]
        echo(f'Mod found: {key} {options[key]["Version"]}')
        options[key]["Name"] = key
        selection = ModDownload(options[key], options[key]["URL"])

    if len(matchKeys) > 1:
        echo("Mods found:")
        idx = 1
        for key in matchKeys:
            if max > -1 and idx > max:
                break
            echo(f'  [{idx}] {key} {options[key]["Version"]}')
            idx += 1

        choice = click.prompt(
            "Select mod to add",
            type=click.IntRange(0, idx),
            default=0,
            show_default=False,
        )
        if choice:
            key = matchKeys[choice - 1]
            echo(f'Selected mod: {key} {options[key]["Version"]}')
            options[key]["Name"] = key
            selection = ModDownload(options[key], options[key]["URL"])

    return selection


def resolve_dependencies(
    mods: t.Iterable[ModMeta], database: t.Optional[t.Dict[str, t.Any]] = None
):
    database = database or get_dependency_graph()
    deps = combined_dependencies(mods, database)
    sorted_deps = sorted(deps.values(), key=lambda dep: dep.Name)

    return sorted_deps


def get_mod_download(mod: str, mod_list: t.Dict[str, t.Any]):
    mod_info = mod_list[mod]
    mod_info["Name"] = mod
    return ModDownload(ModMeta(mod_info), mod_info["URL"], mod_info["MirrorURL"])


def resolve_mods(mods: t.Sequence[str]):
    resolved: t.List[ModDownload] = list()
    unresolved: t.List[str] = list()
    mod_list = get_mod_list()

    for mod in mods:
        parsed_url = urllib.parse.urlparse(mod)
        download = None
        url = None

        # Special case to try to resolve 1-Click install links through mod database
        if parsed_url.scheme == "everest":
            match = re.match(
                "^(https://gamebanana.com/mmdl/.*),.*,.*$", parsed_url.path
            )
            if match:
                gb_url = match[1]
                matches = {
                    key: val for key, val in mod_list.items() if gb_url == val["URL"]
                }
                if len(matches) > 0:
                    download = prompt_mod_selection(matches)
                else:
                    parsed_url = urllib.parse.urlparse(gb_url)

        # Install from local filesystem
        if mod.endswith(".zip") and os.path.exists(mod):
            parsed_url = urllib.parse.ParseResult(
                "file", "", os.path.abspath(mod), "", "", ""
            )
            meta = read_mod_info(parsed_url.path)
            if meta:
                download = ModDownload(meta, urllib.parse.urlunparse(parsed_url))
                echo(f"Mod found: {meta}")

        # Mod ID match
        elif mod in mod_list:
            mod_info = mod_list[mod]
            mod_info["Name"] = mod
            download = ModDownload(
                ModMeta(mod_info), mod_info["URL"], mod_info["MirrorURL"]
            )
            echo(f'Mod found: {mod} {mod_info["Version"]}')

        # GameBanana submission URL
        elif (
            parsed_url.scheme in ("http", "https")
            and parsed_url.netloc == "gamebanana.com"
            and parsed_url.path.startswith("/mods")
            and parsed_url.path.split("/")[-1].isdigit()
        ):
            modID = int(parsed_url.path.split("/")[-1])
            matches = {
                key: val
                for key, val in mod_list.items()
                if modID == val["GameBananaId"]
            }
            if len(matches) > 0:
                download = prompt_mod_selection(matches)
            else:
                downloads = json.load(
                    download_with_progress(
                        f"https://gamebanana.com/apiv5/Mod/{modID}?_csvProperties=_aFiles",
                        None,
                        "Retrieving download list",
                    )
                )["_aFiles"]
                echo("Available downloads:")
                idx = 1
                for d in downloads:
                    echo(f'  [{idx}] {d["_sFile"]} {d["_sDescription"]}')
                    idx += 1

                selection = click.prompt(
                    "Select file to download",
                    type=click.IntRange(0, idx),
                    default=0,
                    show_default=False,
                )
                if selection:
                    d = downloads[selection - 1]
                    echo(f'Selected file: {d["_sFile"]}')
                    url = str(d["_sDownloadUrl"])
                else:
                    echo("Aborted!")

        # Google Drive share URL
        elif (
            parsed_url.scheme in ("http", "https")
            and parsed_url.netloc == "drive.google.com"
            and parsed_url.path.startswith("/file/d/")
        ):
            file_id = parsed_url.path[len("/file/d/") :].split("/")[0]
            url = "https://drive.google.com/uc?export=download&id=" + file_id

        elif parsed_url.scheme and parsed_url.path:
            url = mod

        # Possible GameBanana Submission ID
        elif mod.isdigit():
            modID = int(mod)
            matches = {
                key: val
                for key, val in mod_list.items()
                if modID == val["GameBananaId"]
            }
            if len(matches) > 0:
                download = prompt_mod_selection(matches)

        if download:
            resolved.append(download)
        elif url:
            unresolved.append(url)
        else:
            raise click.ClickException(f"Mod '{mod}' could not be resolved.")

    return resolved, unresolved


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
@click.argument("mods", nargs=-1)
@click.option(
    "--search", is_flag=True, help="Use the Celeste mod search API to find a mod."
)
@click.option("--random", is_flag=True, help="Install a random mod.")
@click.option(
    "--no-deps", is_flag=True, default=False, help="Skip installing dependencies."
)
@click.option("--optional-deps", is_flag=True, default=False, hidden=True)
@clickExt.yes_option()
@clickExt.force_option()
@click.pass_context
def add(
    ctx: click.Context,
    name: Install,
    mods: t.Tuple[str, ...],
    search: bool,
    random: bool,
    no_deps: bool,
    optional_deps: bool,
):
    """Add one or more mods.

    MODS can be one or more of: mod ID, local zip, zip URL, 1-Click install link, Google Drive share link, GameBanana page, or GameBanana submission ID."""
    if random:
        mods = (
            urllib.request.urlopen(
                "https://max480-random-stuff.appspot.com/celeste/random-map"
            ).url,
        )

    if not mods:
        param = next(param for param in ctx.command.params if param.name == "mods")
        raise click.MissingParameter(ctx=ctx, param=param)

    install = name
    mod_folder = os.path.join(os.path.dirname(install.path), "Mods")
    mod_list = get_mod_list()

    installed_list = installed_mods(mod_folder, valid=True, with_size=True)
    installed_list = {
        meta.Name: meta
        for meta in tqdm(
            installed_list, desc="Reading Installed Mods", leave=False, unit=""
        )
    }

    resolved: t.List[ModDownload] = list()
    unresolved: t.List[str] = list()

    def process_zip(zip: str, name: str):
        meta = read_mod_info(zip)
        if meta:
            if meta.Name in installed_list:
                if os.path.isdir(installed_list[meta.Name].Path):
                    raise IsADirectoryError("Could not overwrite non-zipped mod.")
                shutil.move(zip, installed_list[meta.Name].Path)
            else:
                resolved.append(
                    ModDownload(meta, f"file:{urllib.request.pathname2url(zip)}")
                )
        elif clickExt.confirm_ext(
            f"'{name}' does not seem to be an Everest mod.\nInstall anyways?",
            default=False,
            dangerous=True,
        ):
            filename = click.prompt("Save as file")
            if filename:
                filename = filename if filename.endswith(".zip") else filename + ".zip"
                shutil.move(zip, os.path.join(mod_folder, filename))
        else:
            echo(f"Skipped install for {name}.")

    # Query mod search API
    if search:
        mod_list = get_mod_list()
        search_result = search_mods(" ".join(mods))
        matches = {}
        for item in search_result:
            matches.update(
                {
                    mod: data
                    for mod, data in mod_list.items()
                    if data["GameBananaId"] == item["itemid"]
                }
            )

        if len(matches) < 1:
            echo("No results found.")
            exit()

        match = prompt_mod_selection(matches, max=9)
        if not match:
            raise click.Abort()
        resolved = [match]

    # Read from stdin
    elif mods == ("-",):
        with click.open_file("-", mode="rb") as stdin:
            with fs.temporary_file(persist=True) as temp:
                if stdin.isatty():
                    raise TTYError("no input.")
                with click.open_file(temp, mode="wb") as file:
                    read_with_progress(
                        stdin, file, label="Reading from stdin", clear_progress=True  # type: ignore
                    )
                process_zip(temp, "stdin")

    else:
        resolved, unresolved = resolve_mods(mods)

    if len(unresolved) > 0:
        echo("The following mods could not be resolved:")
        for s in unresolved:
            echo(f"\t{s}")
        download_size = sum(get_download_size(url) for url in unresolved)
        echo(
            f"Downloading them all will use up to {format_bytes(download_size)} disk space."
        )
        if clickExt.confirm_ext(
            "Download and attempt to resolve them before continuing?", default=True
        ):
            for url in unresolved:
                with fs.temporary_file(persist=True) as file:
                    download_with_progress(url, file, f"Downloading {url}", clear=True)
                    process_zip(file, url)

            unresolved = []

    if len(resolved) > 0:
        installed, resolved = partition(
            lambda mod: mod.Meta.Name in installed_list, resolved
        )
        if len(installed) > 0:
            echo("Already installed:")
            for mod in installed:
                echo(installed_list[mod.Meta.Name])

    if len(resolved) > 0:
        if not no_deps:
            echo("Resolving dependencies...")
            dependencies = resolve_dependencies(map(lambda d: d.Meta, resolved))
            special, dependencies = partition(
                lambda mod: mod.Name in ("Celeste", "Everest"), dependencies
            )
            deps_install: t.List[t.Any] = []
            deps_update: t.List[UpdateInfo] = []
            deps_blacklisted: t.List[ModMeta] = []
            for dep in dependencies:
                if dep.Name in installed_list:
                    installed_mod = installed_list[dep.Name]
                    if installed_mod.Version.satisfies(dep.Version):
                        if installed_mod.Blacklisted:
                            deps_blacklisted.append(installed_mod)
                    else:
                        deps_update.append(
                            UpdateInfo(
                                installed_mod, dep.Version, mod_list[dep.Name]["URL"]
                            )
                        )
                else:
                    deps_install.append(dep)

            deps_install, unregistered = partition(
                lambda mod: mod.Name in mod_list, deps_install
            )
            if len(unregistered) > 0:
                echo(f"{len(unregistered)} dependencies could not be found:")
                for mod in unregistered:
                    echo(mod)
                if not clickExt.confirm_ext(
                    "Install without missing dependencies?",
                    default=True,
                    dangerous=True,
                ):
                    raise click.Abort()

            deps_install = [
                get_mod_download(mod.Name, mod_list) for mod in deps_install
            ]
        else:
            deps_install = deps_update = deps_blacklisted = []
        count = len(deps_install) + len(resolved)
        if count > 0:
            echo(f"\t{count} To Install:")
            for download in itertools.chain(deps_install, resolved):
                echo(download.Meta)
        count = len(deps_update)
        if count > 0:
            echo(f"\t{count} To Update:")
            for mod in deps_update:
                echo(f"{mod.Old} -> {mod.New}")
        count = len(deps_blacklisted)
        if count > 0:
            echo(f"\t{count} To Enable:")
            for mod in deps_blacklisted:
                echo(mod)

        # Hack to allow assigning to a variable out of scope
        download_size_ref = [0]

        def download_size_key(mod: t.Union[UpdateInfo, ModDownload]):
            size = (
                get_download_size(mod.Url, mod.Old.Size)
                if isinstance(mod, UpdateInfo)
                else mod.Meta.Size
            )
            download_size_ref[0] += size
            return size

        sorted_dep_downloads = sorted(
            itertools.chain(deps_install, deps_update), key=download_size_key
        )
        sorted_main_downloads = sorted(resolved, key=download_size_key)

        download_size = download_size_ref[0]
        if download_size >= 0:
            echo(
                f"After this operation, an additional {format_bytes(download_size)} disk space will be used"
            )
        else:
            echo(
                f"After this operation, {format_bytes(abs(download_size))} disk space will be freed"
            )

        if not clickExt.confirm_ext("Continue?", default=True):
            raise click.Abort()

        with timed_progress("Downloaded files in {time:.3f} seconds."):
            enable_mods(
                mod_folder,
                *itertools.chain(
                    (os.path.basename(mod.Path) for mod in deps_blacklisted),
                    (os.path.basename(mod.Old.Path) for mod in deps_update),
                    # blindly attempt to enable any other mods that will be downloaded
                    (f"{mod.Meta.Name}.zip" for mod in deps_install),
                ),
            )

            download_threaded(
                mod_folder,
                sorted_dep_downloads,
                late_downloads=sorted_main_downloads,
                thread_count=10,
            )

        if no_deps:
            exit()

        everest_min = next(
            (dep.Version for dep in special if dep.Name == "Everest"), Version(1, 0, 0)  # type: ignore
        )
        current_everest = Version(
            1, int(install.get_cache().get("everestbuild", "0")), 0
        )
        if not current_everest.satisfies(everest_min):
            echo(
                f"Installed Everest ({current_everest}) does not satisfy minimum requirement ({everest_min})."
            )
            if clickExt.confirm_ext("Update Everest?", default=True):
                mons_cli.main(args=["install", install.name, str(everest_min)])


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
@click.argument("mods", nargs=-1, required=True)
@click.option(
    "-r",
    "--recurse",
    is_flag=True,
    help="Remove all exclusive dependencies recursively.",
)
@clickExt.force_option()
def remove(name: Install, mods: t.List[str], recurse: bool):
    """Remove installed mods."""
    mod_folder = os.path.join(os.path.dirname(name.path), "Mods")
    installed_list = installed_mods(mod_folder, valid=True, with_size=True)
    installed_list = {
        meta.Name: meta
        for meta in tqdm(
            installed_list, desc="Reading Installed Mods", leave=False, unit=""
        )
    }

    resolved, unresolved = partition(lambda mod: mod in installed_list, mods)

    if len(unresolved) > 0:
        echo("The following mods could not be found, and will not be removed:")
        for mod in unresolved:
            echo(f"\t{mod}")
        if not clickExt.confirm_ext("Continue anyways?", default=False, dangerous=True):
            raise click.Abort()

    metas = [installed_list[mod] for mod in resolved]

    if len(metas) < 1:
        echo("No mods to remove.")
        exit()

    echo(str(len(metas)) + " mods will be removed:")
    for mod in metas:
        echo(f"\t{mod}")

    removable = []
    if recurse:
        all_dependencies = resolve_dependencies(
            [meta for meta in installed_list.values() if not meta.Name in mods],
            installed_list,
        )

        dependencies = resolve_dependencies(
            metas, {mod: meta for mod, meta in installed_list.items() if mod in mods}
        )

        removable = {dep.Name for dep in dependencies}.difference(
            {dep.Name for dep in all_dependencies}
        )
        removable = [installed_list[dep] for dep in removable if dep in installed_list]

        echo(str(len(removable)) + " dependencies will also be removed:")
        for dep in removable:
            echo(f"\t{dep}")

    total_size = sum(mod.Size for mod in itertools.chain(metas, removable))
    echo(f"After this operation, {format_bytes(total_size)} disk space will be freed.")

    if not clickExt.confirm_ext("Remove mods?", default=True, dangerous=True):
        raise click.Abort()

    folders: t.List[ModMeta] = []
    with click.progressbar(
        label="Deleting files", length=len(removable) + len(metas)
    ) as progress:
        for mod in itertools.chain(removable, metas):
            # This should be handled by catching IsADirectoryError but for some reason it raises PermissionError instead so...
            if os.path.isdir(mod.Path):
                folders.append(mod)
            else:
                os.remove(mod.Path)
            progress.update(1)

    if len(folders) > 0:
        echo("The following unzipped mods were not removed:")
        for mod in folders:
            echo(f"\t{mod} ({os.path.basename(mod.Path)}/)")


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
# @click.argument('mod', required=False)
@click.option("--all", is_flag=True, help="Update all installed mods.", required=True)
@click.option(
    "--enabled", is_flag=True, help="Update currently enabled mods.", default=None
)
@click.option(
    "--upgrade-only", is_flag=True, help="Only update if new file has a higher version."
)
@clickExt.yes_option()
def update(name: Install, all: bool, enabled: t.Optional[bool], upgrade_only: bool):
    """Update installed mods."""
    if not all:
        raise click.UsageError(
            "this command can currently only be used with the --all option"
        )

    mod_list = get_mod_list()
    updates: t.List[UpdateInfo] = []
    has_updates = False
    total_size: int = 0
    if all:
        mods_folder = os.path.join(os.path.dirname(name.path), "Mods")
        installed = installed_mods(
            mods_folder,
            blacklisted=invert(enabled),
            dirs=False,
            valid=True,
            with_size=True,
            with_hash=True,
        )
        updater_blacklist = os.path.join(mods_folder, "updaterblacklist.txt")
        updater_blacklist = os.path.exists(updater_blacklist) and read_blacklist(
            updater_blacklist
        )
        for meta in installed:
            if meta.Name in mod_list and (
                not updater_blacklist
                or os.path.basename(meta.Path) not in updater_blacklist
            ):
                server = mod_list[meta.Name]
                latest_hash = server["xxHash"][0]
                latest_version = Version.parse(server["Version"])
                if (
                    meta.Hash
                    and latest_hash != meta.Hash
                    and (not upgrade_only or latest_version > meta.Version)
                ):
                    update = UpdateInfo(
                        meta,
                        latest_version,
                        server["URL"],
                    )
                    if not has_updates:
                        echo("Updates available:")
                        has_updates = True
                    echo(f"\t{update.Old} -> {update.New}")
                    updates.append(update)
                    total_size += server["Size"] - update.Old.Size

    if not has_updates:
        echo("All mods up to date")
        return

    echo(
        ngettext(
            f"{len(updates)} update found",
            f"{len(updates)} updates found",
            len(updates),
        )
    )

    if total_size >= 0:
        echo(
            f"After this operation, an additional {format_bytes(total_size)} disk space will be used"
        )
    else:
        echo(
            f"After this operation, {format_bytes(abs(total_size))} disk space will be freed"
        )

    if not clickExt.confirm_ext("Continue?", default=True):
        raise click.Abort()

    with timed_progress("Downloaded files in {time:.3f} seconds."):
        download_threaded("", updates, thread_count=10)


@cli.command(no_args_is_help=True, cls=clickExt.CommandExt)
@clickExt.install("name")
# @click.argument('mod', required=False)
@click.option("--all", is_flag=True, help="Resolve all installed mods.", required=True)
@click.option(
    "--enabled", is_flag=True, help="Resolve currently enabled mods.", default=None
)
@click.option("--no-update", is_flag=True, help="Don't update outdated dependencies.")
@clickExt.yes_option()
def resolve(name: Install, all: bool, enabled: t.Optional[bool], no_update: bool):
    """Resolve any missing or outdated dependencies."""
    if not all:
        raise click.UsageError(
            "this command can currently only be used with the --all option"
        )

    install = name

    mods_folder = os.path.join(os.path.dirname(install.path), "Mods")
    installed = installed_mods(mods_folder, valid=True, blacklisted=invert(enabled))
    installed = list(
        tqdm(installed, desc="Reading Installed Mods", leave=False, unit="")
    )
    installed_dict = {meta.Name: meta for meta in installed}

    deps = resolve_dependencies(installed)

    special, deps_installed, deps_missing = multi_partition(
        lambda meta: meta.Name in ("Celeste", "Everest"),
        lambda meta: meta.Name in installed_dict,
        iterable=deps,
    )

    deps_outdated = (
        [
            dep
            for dep in deps_installed
            if not installed_dict[dep.Name].Version.satisfies(dep.Version)
        ]
        if not no_update
        else []
    )

    if len(deps_missing) + len(deps_outdated) < 1:
        return

    echo(
        f'{len(deps_missing) + len(deps_outdated)} dependencies missing{" or outdated" if len(deps_outdated) else ""}, attempting to resolve...'
    )

    mod_list = get_mod_list()
    deps_install = [
        get_mod_download(mod.Name, mod_list)
        for mod in deps_missing
        if mod.Name in mod_list
    ]
    deps_update = [
        UpdateInfo(installed_dict[dep.Name], dep.Version, mod_list[dep.Name]["URL"])
        for dep in deps_outdated
        if dep.Name in mod_list
    ]

    unresolved = (len(deps_missing) + len(deps_outdated)) - (
        len(deps_install) + len(deps_update)
    )
    if unresolved != 0:
        echo(f"{unresolved} mods could not be resolved.")

    if len(deps_install) > 0:
        echo("\tTo Install:")
    for mod in deps_install:
        echo(mod.Meta)
    if len(deps_update) > 0:
        echo("\tTo Update:")
    for mod in deps_update:
        echo(f"{mod.Old} -> {mod.New}")

    download_size_ref = [0]

    def download_size_key(mod: t.Union[UpdateInfo, ModDownload]):
        size = (
            get_download_size(mod.Url, mod.Old.Size)
            if isinstance(mod, UpdateInfo)
            else mod.Meta.Size
        )
        download_size_ref[0] += size
        return size

    sorted_dep_downloads = sorted(
        itertools.chain(deps_install, deps_update), key=download_size_key
    )

    download_size = download_size_ref[0]
    if download_size >= 0:
        echo(
            f"After this operation, an additional {format_bytes(download_size)} disk space will be used"
        )
    else:
        echo(
            f"After this operation, {format_bytes(abs(download_size))} disk space will be freed"
        )

    if not clickExt.confirm_ext("Continue?", default=True):
        raise click.Abort

    with timed_progress("Downloaded files in {time:.3f} seconds."):
        download_threaded(mods_folder, sorted_dep_downloads, thread_count=10)

    everest_min = next(
        (dep.Version for dep in special if dep.Name == "Everest"), Version(1, 0, 0)
    )
    current_everest = Version(1, int(install.get_cache().get("everestbuild", "0")), 0)
    if not current_everest.satisfies(everest_min):
        echo(
            f"Installed Everest ({current_everest}) does not satisfy minimum requirement ({everest_min})."
        )
        if clickExt.confirm_ext("Update Everest?", default=True):
            mons_cli.main(args=["install", install.name, str(everest_min)])



import atexit
import os
import shutil
import tempfile
import typing as t
from contextlib import contextmanager

from mons.baseUtils import tryExec


def find_file(path: str, files: t.Iterable[str]):
    """Return the first file in :param:`files` found in :param:`path`, or :literal:`None`."""
    for file in files:
        if os.path.isfile(os.path.join(path, file)):
            return file
    return None


# shutils.copytree(dirs_exist_ok) replacement https://stackoverflow.com/a/15824216
def copy_recursive_force(src: str, dest: str, ignore=None):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        if ignore is not None:
            ignored = ignore(src, files)
        else:
            ignored: t.Collection[str] = set()
        for f in files:
            if f not in ignored:
                copy_recursive_force(
                    os.path.join(src, f), os.path.join(dest, f), ignore
                )
    else:
        shutil.copyfile(src, dest)


def folder_size(path):
    """Compute the size of a folder on disk as reported by :func:`os.stat`."""
    total_size = 0
    for dirpath, _, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


def is_unchanged(src: str, dest: str):
    """Returns :literal:`True` if :param:`src` has not been changed after :param:`dest` was."""
    if os.path.exists(dest):
        return os.stat(dest).st_mtime - os.stat(src).st_mtime >= 0
    return False


@contextmanager
def relocated_file(src: str, dest: str):
    """Temporarily moves :param:`src` to :param:`dest`."""
    file = shutil.move(src, dest)
    try:
        yield file
    finally:
        shutil.move(file, src)


@contextmanager
def copied_file(src: str, dest: str):
    """Temporarily copies :param:`src` to :param:`dest`."""
    file = shutil.copy(src, dest)
    try:
        yield file
    finally:
        os.remove(file)


@contextmanager
def temporary_file(persist=False):
    fd, path = tempfile.mkstemp(suffix="_mons")
    if persist:
        atexit.register(tryExec, os.remove, path)
    os.close(fd)
    try:
        yield path
    finally:
        if not persist and os.path.isfile(path):
            os.remove(path)



import os
import shutil
import urllib.request
import zipfile

import pytest

from mons.mons import cli as mons_cli

TEST_INSTALL = "_mons_testing"
GITHUB_REPO = "https://github.com/EverestAPI/Everest"


def setup_install(runner, test_install):
    if runner.invoke(mons_cli, ["add", TEST_INSTALL, test_install]).exit_code != 0:
        pytest.skip()


def test_install(runner, test_install, cache: pytest.Cache):
    setup_install(runner, test_install)

    # install preferred branch (defaults to stable)
    result = runner.invoke(mons_cli, ["install", TEST_INSTALL])
    assert result.exit_code == 0, result.output
    assert "Install success" in result.output


def try_get_artifact(test_install):
    install_dir = os.path.dirname(test_install)
    if os.path.exists(os.path.join(install_dir, "olympus-build.zip")):
        return os.path.join(install_dir, "olympus-build.zip")
    elif os.path.exists(os.path.join(install_dir, "main.zip")):
        return os.path.join(install_dir, "main.zip")
    else:
        pytest.skip("Everest build artifact not available")


def test_install_url(runner, test_install):
    url = "file://" + try_get_artifact(test_install)

    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["install", TEST_INSTALL, "--url", url])
    assert result.exit_code == 0, result.output
    assert "Install success" in result.output


def test_install_zip(runner, test_install):
    url = try_get_artifact(test_install)

    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["install", TEST_INSTALL, "--zip", url])
    assert result.exit_code == 0, result.output
    assert "Install success" in result.output


@pytest.mark.xfail(
    not (shutil.which("dotnet") or shutil.which("msbuild")),
    reason="no .NET build tool found",
    run=False,
)
def test_install_src(runner, test_install, tmp_path):
    dest = os.path.join(tmp_path, "stable.zip")
    urllib.request.urlretrieve(f"{GITHUB_REPO}/archive/refs/heads/stable.zip", dest)

    with zipfile.ZipFile(dest, "r") as zip:
        zip.extractall(tmp_path)

    source_dir = os.path.join(tmp_path, "Everest-stable")
    assert os.listdir(source_dir), "Failed to download source artifact: " + source_dir

    setup_install(runner, test_install)

    result = runner.invoke(mons_cli, ["install", TEST_INSTALL, f"--src={source_dir}"])
    assert result.exit_code == 0, result.output
    assert "Install success" in result.output



from __future__ import annotations  # ABCs are not generic prior to 3.9

import os
import typing as t
from collections.abc import MutableMapping
from configparser import SectionProxy

from mons.utils import getMD5Hash
from mons.utils import parseExeInfo
from mons.utils import VANILLA_HASH


class Install:

    DEFAULTS: t.Dict[str, str]

    def __init__(
        self,
        name: str,
        path: str,
        cache: t.Optional[MutableMapping[str, str]] = None,
        data: t.Optional[MutableMapping[str, str]] = None,
    ) -> None:
        self.name = name
        assert os.path.exists(path)
        self.path = path
        self.cache: MutableMapping[str, str] = cache or dict()
        self.data: MutableMapping[str, str] = data or dict()

    def update_cache(self, read_exe=True):
        hash = getMD5Hash(self.path)
        if self.cache.get("Hash", "") == hash:
            # Cache is up to date
            return

        self.cache["Hash"] = hash
        self.cache["Everest"] = str(True)
        self.cache["CelesteVersion"] = ""

        version = VANILLA_HASH.get(hash, None)
        if version:
            self.cache["CelesteVersion"], self.cache["Framework"] = version
            self.cache["Everest"] = str(False)

        else:
            orig_path = os.path.join(os.path.dirname(self.path), "orig", "Celeste.exe")
            if os.path.isfile(orig_path):
                orig_hash = getMD5Hash(orig_path)
                version = VANILLA_HASH.get(orig_hash, None)
                if version:
                    self.cache["CelesteVersion"], self.cache["Framework"] = version
                    # If orig_hash is known but hash is not, then Celeste.exe has been modified
                    self.cache["Everest"] = str(True)

        if read_exe and self.cache["Everest"]:
            (
                self.cache["Everest"],
                self.cache["EverestBuild"],
                self.cache["Framework"],
            ) = map(str, parseExeInfo(self.path))

    def version_string(self):
        self.update_cache(read_exe=False)

        versionStr: str = self.cache.get("CelesteVersion", "unknown")
        framework: t.Optional[str] = self.cache.get("Framework", None)
        if framework:
            versionStr += f"-{framework.lower()}"
        everestBuild: t.Optional[str] = self.cache.get("EverestBuild", None)
        if everestBuild:
            versionStr += f" + 1.{everestBuild}.0"
        elif str(self.cache.get("Everest", None)) == "True":
            versionStr += " + Everest(unknown version)"
        return versionStr

    def get_cache(self):
        self.update_cache()
        return self.cache

    def __setitem__(self, name: str, value: str):
        self.data[name] = value

    def __getitem__(self, name: str):
        return self.data.get(name, "") or Install.DEFAULTS.get(name, "")

    def serialize(self):
        return {**self.data, "path": self.path}



import os
import pathlib

import pytest
from click.testing import CliRunner

from mons.mons import cli as mons_cli


def pytest_collection_modifyitems(session, config, items):
    module = pathlib.Path(os.path.dirname(__file__))
    for item in items:
        if module == item.path.parent:
            item.add_marker(pytest.mark.integration_test)


def get_commands(cli, *, prefix=""):
    for cmd in cli.commands.values():
        if hasattr(cmd, "commands"):
            yield from get_commands(cmd, prefix=prefix + f"{cmd.name} ")
        else:
            cmd.qualified_name = prefix + cmd.name
            yield cmd


@pytest.fixture(
    scope="module",
    params=list(get_commands(mons_cli)),
    ids=lambda cmd: cmd.qualified_name,
)
def command(request):
    yield request.param


@pytest.fixture(scope="function")
def runner():
    """
    runner.isolated_filesystem isn't necessary because mons never operates directly on the working directory
    but if it did:

    ```
    with runner.isolated_filesystem(temp_dir=tmp_path):
        yield runner
    ```
    """
    return CliRunner()



import operator as op

import pytest

from mons.version import Version


@pytest.mark.parametrize(
    ("version", "expect"),
    [
        ("NoVersion", Version(1, 0)),
        ("1.2.3.4", Version(1, 2, 3, 4)),
        ("1.2.3-pre", Version(1, 2, 3)),
    ],
)
def test_version_parse(version, expect):
    assert Version.parse(version) == expect


@pytest.mark.parametrize(
    ("version", "required", "expect"),
    [
        (Version(1, 2, 3, 4), Version(1, 2, 3, 4), True),
        (Version(1, 2, 3, 4), Version(1, 2, 3, 3), False),
        (Version(1, 2, 3, 4), Version(1, 3, 2, 4), True),
        (Version(1, 2, 3, 4), Version(2, 2, 3, 4), False),
    ],
)
def test_version_satisfies(version, required, expect):
    assert required.satisfies(version) == expect


@pytest.mark.parametrize(
    ("left", "right", "comparison"),
    [
        (Version(1, 2, 3, 4), Version(1, 2, 3, 4), op.eq),
        (Version(1, 2, 3, 4), Version(1, 0, 0, 0), op.ne),
        (Version(1, 2, 3, 4), Version(1, 2, 3, 3), op.gt),
        (Version(1, 2, 3, 4), Version(1, 3, 3, 4), op.lt),
    ],
)
def test_version_compare(left, right, comparison):
    assert comparison(left, right)



import sys

import click

from mons.mons import cli

if sys.version_info >= (3, 8):  # novermin
    from importlib.metadata import version

    version = version("mons")
else:
    import pkg_resources

    version = pkg_resources.get_distribution("mons").version

# Override built in version detection to fix issues when running as __main__
click.version_option(version=version, package_name="mons")(cli)

cli()



import pytest

from mons.formatting import *


@pytest.mark.parametrize("input, expected", [(1299999, "1.2 MiB")])
def test_format_bytes(input, expected):
    assert format_bytes(input) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            {"foo": "value", "bar": "value"},
            """
foo\tvalue
bar\tvalue
""".strip(),
        ),
        (
            {"foo": "value", "foobarbaz": "value"},
            """
foo      \tvalue
foobarbaz\tvalue
""".strip(),
        ),
        ({}, ""),
    ],
)
def test_format_columns(input, expected):
    assert format_columns(input) == expected



#!/usr/bin/env python
import os
import typing as t
from importlib import import_module

import click

import mons.clickExt as clickExt
from mons.config import Env
from mons.config import UserInfo


@click.group(cls=clickExt.CatchErrorsGroup)
@click.pass_context
@click.version_option()
def cli(ctx: click.Context):
    ctx.obj = ctx.with_resource(UserInfo())

    # Inject another context as the parent
    env_ctx = click.Context(ctx.command, ctx.parent, obj=Env())
    ctx.parent = env_ctx


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("command", nargs=-1)
@click.pass_context
def help(ctx: click.Context, command: t.List[str]):
    """Display help text for a command"""
    # No args means print program help
    if len(command) < 1 and ctx.parent:
        click.echo(cli.get_help(ctx.parent))
        exit(0)

    group = cli
    cmd_path = []
    for cmd_name in command:
        cmd_path.append(cmd_name)
        cmd = group.get_command(ctx, cmd_name)
        if not cmd:
            err_msg = "No such command '{}.'".format(" ".join(cmd_path))
            click.echo(str(click.BadArgumentUsage(err_msg, ctx)))
            exit(click.BadArgumentUsage.exit_code)

        if isinstance(cmd, click.Group):
            group = cmd
            continue
        else:
            # ctx currently thinks it's for the help command, this corrects the usage text
            ctx.info_name = " ".join(cmd_path)
            click.echo(cmd.get_help(ctx))
            exit(0)


cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))
for filename in os.listdir(cmd_folder):
    if filename.endswith(".py") and not filename.startswith("__"):
        import_module(f"mons.commands.{filename[:-3]}")
